from fsgs.amiga.roms import *


class Amiga(object):
    # FIXME: setting fake checksum, for now
    INTERNAL_ROM_SHA1 = "0000000000000000000000000000000000000000"

    MAX_FLOPPY_DRIVES = 4
    MAX_FLOPPY_IMAGES = 20
    MAX_CDROM_DRIVES = 4
    MAX_CDROM_IMAGES = 20
    MAX_HARD_DRIVES = 4

    models = [
        {
            "id": "A1000",
            "title": "A1000",
            "subtitle": "1.2 ROM, 512 KB Chip RAM",
            "cd_based": False,
            "kickstarts": A1000_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A500/512K",
            "title": "Amiga 500 (512K)",
            # "subtitle": "1.3 ROM, 512 KB Chip RAM",
            "subtitle": "1.3 ROM",
            "cd_based": False,
            "kickstarts": A500_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A500",
            "title": "A500",
            # "subtitle": "1.3 ROM, 512 KB Chip  + 512 KB Slow RAM",
            # "subtitle": "1.3 ROM, 512 KB RAM Trapdoor Expansion",
            "subtitle": '1.3 ROM, 512 KB "Slow" RAM',
            # "subtitle": "1.3 ROM, 512 KB  RAM",
            "cd_based": False,
            "kickstarts": A500_KICKSTARTS,
            # "kickstarts": ["0123456789012345678901234567890123456789"],
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A500+",
            "title": "A500+",
            "subtitle": "2.04 ROM",
            "cd_based": False,
            "kickstarts": A500P_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A600",
            "title": "A600",
            "subtitle": "2.05 ROM",
            "cd_based": False,
            "kickstarts": A600_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200/3.0",
            "title": "A1200",
            "subtitle": "3.0 ROM",
            "cd_based": False,
            "kickstarts": A1200_3_0_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200",
            "title": "A1200",
            "subtitle": "3.1 ROM",
            "cd_based": False,
            "kickstarts": A1200_3_1_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200/020",
            "title": "Amiga 1200 (68020)",
            "subtitle": "3.1 ROM, 68020 CPU",
            "cd_based": False,
            "kickstarts": A1200_3_1_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200/1230",
            "title": "A1200",
            "subtitle": "3.1 ROM, Blizzard 1230 IV",
            "cd_based": False,
            "kickstarts": A1200_3_1_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200/1240",
            "title": "A1200",
            "subtitle": "3.1 ROM, Blizzard 1240",
            "cd_based": False,
            "kickstarts": A1200_3_1_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200/1260",
            "title": "A1200",
            "subtitle": "3.1 ROM, Blizzard 1260",
            "cd_based": False,
            "kickstarts": A1200_3_1_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A1200/PPC",
            "title": "A1200",
            "subtitle": "3.1 ROM, Blizzard PPC",
            "cd_based": False,
            "kickstarts": A1200_3_1_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A3000",
            "title": "A3000",
            "subtitle": "3.1 ROM, 2 MB Chip + 8 MB Fast RAM",
            "cd_based": False,
            "kickstarts": A3000_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A4000",
            "title": "A4000",
            "subtitle": "3.1 ROM",
            "cd_based": False,
            "kickstarts": A4000_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A4000/040",
            "title": "Amiga 4000 (68040)",
            "subtitle": "3.1 ROM, 68040 CPU",
            "cd_based": False,
            "kickstarts": A4000_KICKSTARTS,
            "ext_roms": [],
            "defaults": {},
        },
        {
            "id": "A4000/PPC",
            "title": "Amiga 4000 (PPC)",
            "subtitle": "3.1 ROM, CyberStorm PPC",
            "cd_based": False,
            "kickstarts": A4000_KICKSTARTS,
            "ext_roms": [],
            "defaults": {"accelerator": "cyberstorm-ppc"},
        },
        {
            "id": "A4000/OS4",
            "title": "Amiga 4000 (PPC / OS4)",
            "subtitle": "3.1 ROM, CyberStorm PPC (AmigaOS 4.x)",
            "cd_based": False,
            "kickstarts": A4000_KICKSTARTS,
            "ext_roms": [],
            "defaults": {
                "accelerator": "cyberstorm-ppc",
                "graphics_card": "picasso-iv-z3",
            },
        },
        {
            "id": "CD32",
            "title": "CD32",
            "subtitle": "3.1 ROM",
            "cd_based": "CD32",
            "kickstarts": CD32_KICKSTARTS,
            "ext_roms": CD32_EXT_ROMS,
            "defaults": {},
        },
        {
            "id": "CD32/FMV",
            "title": "CD32 + FMV ROM",
            "subtitle": "3.1 ROM, FMV Expansion (ROM)",
            "cd_based": "CD32",
            "kickstarts": CD32_KICKSTARTS,
            "ext_roms": CD32_EXT_ROMS,
            "defaults": {},
        },
        {
            "id": "CDTV",
            "title": "CDTV",
            "subtitle": "1.3 ROM",
            "cd_based": "CDTV",
            "kickstarts": CDTV_KICKSTARTS,
            "ext_roms": CDTV_EXT_ROMS,
            "defaults": {},
        },
    ]

    models_config = [model["id"] for model in models]

    @classmethod
    def is_cd_based(cls, config):
        return cls.get_current_config(config)["cd_based"]

    @classmethod
    def get_current_config(cls, config):
        return cls.get_model_config(config.get("amiga_model"))

    @classmethod
    def get_model_config(cls, model):
        model = model.upper()
        if not model:
            model = "A500"
        for i in range(len(cls.models_config)):
            if cls.models_config[i].upper() == model:
                return cls.models[i]
        return {
            "title": "Dummy",
            "cd_based": False,
            "kickstarts": [],
            "ext_roms": [],
            "defaults": {
                "chip_memory": "0",
                "slow_memory": "0",
                "fast_memory": "0",
                "zorro_iii_memory": "0",
            },
        }



from fsbc import settings
from fsgs import openretro
from fsgs.option import Option
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.ui.IconButton import IconButton


class NewButton(IconButton):
    def __init__(self, parent):
        super().__init__(parent, "new_button.png")
        self.set_tooltip(gettext("New Config"))

    def on_activate(self):
        self.new_config()

    @staticmethod
    def new_config():
        if openretro or settings.get(Option.PLATFORMS_FEATURE):
            platform_id = LauncherConfig.get(Option.PLATFORM)
        else:
            platform_id = None
        LauncherConfig.load_default_config(platform=platform_id)
        # Settings.set("config_changed", "1")
        LauncherSettings.set("parent_uuid", "")



import hashlib
import os

from nose.tools import *

import fstd.adffile
import fstd.mypy


def get_transplant_adf() -> fstd.adffile.ADFFile:
    adf = fstd.adffile.ADFFile(
        os.path.join(os.path.dirname(__file__), "adffile", "transplant.adf")
    )
    return adf


def test_namelist():
    adf = get_transplant_adf()
    expected_names = [
        "devs/",
        "devs/system-configuration",
        "l/",
        "l/disk-validator",
        "loader",
        "mainpart",
        "s/",
        "s/startup-sequence",
    ]
    assert_equal(sorted(adf.namelist()), expected_names)


def test_read_startup_sequence():
    adf = get_transplant_adf()
    data = adf.read("s/startup-sequence")
    assert_equal(data, b"loader\n")


def test_read_mainpart():
    adf = get_transplant_adf()
    data = adf.read("mainpart")
    data_sha1 = hashlib.sha1(data).hexdigest()
    assert_equal(data_sha1, "ef467af52c8a886a6bca6ceb7f263fb8cbee2d59")


def test_mypy():
    fstd.mypy.check_module(fstd.adffile.__name__)



import weakref


class StrWithExplicit(str):
    # def __init__(self, object):
    #     super().__init__(object)
    #     self.explicit = ""

    @property
    def int(self):
        return int(self or "0")


class ImplicitConfig:
    def __init__(self, config, settings):
        self._values = {}
        self._config = config
        self._settings = settings

        # self.__setattr__ = self.__setattr__function

    def items(self):
        return self._values.items()

    def get(self, key, default=""):
        value = self._values.get(key, default)
        explicit = self._config[key]
        # print(explicit)
        if not explicit:
            explicit = self._settings[key]
        if value:
            result = StrWithExplicit(value)
        else:
            result = StrWithExplicit(explicit)
        result.explicit = explicit
        return result

    def __getitem__(self, item):
        result = self.get(item)
        # print("get", item, repr(result))
        return result

    def __setitem__(self, key, value):
        # print("setitem", key, value)
        if not value:
            value = ""
        else:
            value = str(value)
        self._values[key] = value

    def __getattr__(self, item):
        result = self.get(item)
        # print("get", item, repr(result))
        return result
        # print("getattr", item)
        # if item == "_values":
        #     return getattr(self, "_values")
        # return self.get(item)

    def __setattr__(self, key, value):
        if key.startswith("_"):
            self.__dict__[key] = value
        else:
            self[key] = value


class NoParent:
    def __call__(self):
        return None

    # def __nonzero__(self):
    #     return False


class Item:
    def __init__(self, text, active=True):
        self.text = text
        self.extra = ""
        # self.parent = parent
        self._active = active
        self.represents = []
        self.parent = NoParent()
        self.children = []

    # def set_parent(self, item):
    #     self.parent = item

    @property
    def active(self):
        return self._active

    def add(self, child):
        child.parent = weakref.ref(self)
        self.children.append(child)

    def __str__(self):
        return " + {0}".format(self.text)


class ContainerItem(Item):
    """This item is only active if it has children."""

    @property
    def active(self):
        # print("active???", self, len(self.children))
        # return True
        return len(self.children) > 0


class InactiveItem(Item):
    def __init__(self, text):
        super().__init__(text, active=False)

    def __str__(self):
        return " ( {0} )".format(self.text)


class Model:
    def __init__(self, show_all=False):
        self.items = []
        self.show_all = show_all

    def add(self, item):
        # if item.active or self.show_all:
        if True:
            if item.parent:
                for i in range(len(self.items) - 1, -1, -1):
                    if (
                        self.items[0].parent == item.parent
                        or self.items[0] == item.parent
                    ):
                        self.items.insert(i + 1, item)
                        break
                else:
                    self.items.append(item)
            else:
                self.items.append(item)
        return item

    def remove(self, item):
        self.items.remove(item)
        return item

    def last_item(self):
        return self.items[-1]


def normalize(value):
    n = ""
    for c in value.lower():
        if c in "abcdefghijklmnopqrstuvwxyz0123456789+":
            n += c
    return n


def create_joystick_port_item(c, num: int) -> Item:
    # FIXME: retrieve uae_ option
    mode = c.get("joystick_port_{0}_mode".format(num))
    if mode == "nothing":
        item = InactiveItem("Joystick Port {0}".format(num))
    else:
        item = Item("Joystick Port {0}".format(num))
        # if mode == "joystick":
        #     device_text = "[J]"
        # elif mode == "mouse":
        #     device_text = "[M]"
        # elif mode == "cd32_gamepad":
        #     device_text = "[C]"
        # else:
        #     device_text = "[?]"
        # # FIXME: auto-calculated device
        # device = c.get("joystick_port_{0}".format(num))
        # if not device:
        #     device = "???"
        # device_text += " " + device
        # device_item = Item(device_text)
        # item.add(device_item)
    return item


def create_slirp_item(c):
    implementation = c.uae_slirp_implementation
    if implementation == "auto":
        name = "Auto"
    elif implementation == "none":
        name = "None"
    elif implementation == "builtin":
        name = "Built-in"
    elif implementation == "qemu":
        name = "QEMU"
    else:
        name = implementation
    slirp_item = Item("Slirp ({})".format(name))

    if c.uae_slirp_ports:
        slirp_ports_item = Item("Ports: " + c.uae_slirp_ports)
        slirp_item.add(slirp_ports_item)
    if c.uae_slirp_redir:
        slirp_redirect_item = Item("Redirect: " + c.uae_slirp_redir)
        slirp_item.add(slirp_redirect_item)

    return slirp_item


def create_model(c, show_all=False):
    model = Model(show_all=show_all)

    # if c.ntsc_mode:
    #     model_item = Item(c.int_model_name + " NTSC")
    # else:
    #     model_item = Item(c.int_model_name + " PAL")
    # model_item.represents = ["amiga_model", "ntsc_mode"]
    # model.add(model_item)

    text = c.int_chipset_name
    if c.uae_chipset_compatible != "-":
        text += " ({0})".format(c.uae_chipset_compatible)
    if c.ntsc_mode:
        text += " NTSC"
    else:
        text += " PAL"
    chipset_item = Item(text)
    chipset_item.represents = [
        "uae_chipset",
        "uae_chipset_compatible",
        "ntsc_mode",
    ]
    model.add(chipset_item)

    if c.uae_chipset == "aga":
        text = "Alice (AGA)"
        if c.ntsc_mode:
            text += " NTSC"
        else:
            text += " PAL"
        # alice_item = Item(text)
        # chipset_item.add(alice_item)
        # lisa_item = Item("Lisa (AGA)")
        # chipset_item.add(lisa_item)
        # paula_item = Item("Paula (AGA)")
        # chipset_item.add(paula_item)
    else:
        # http://en.wikipedia.org/wiki/MOS_Technology_Agnus
        if c.uae_chipset in ["ecs_agnus", "ecs"]:
            text = "Agnus (ECS)"
        else:
            text = "Agnus"
        if c.ntsc_mode:
            text += " NTSC"
        else:
            text += " PAL"
        # agnus_item = Item(text)
        # chipset_item.add(agnus_item)
        # if c.uae_chipset in ["ecs_denise", "ecs"]:
        #     denise_item = Item("Denise (ECS)")
        # else:
        #     denise_item = Item("Denise")
        # chipset_item.add(denise_item)
        # if c.uae_chipset in ["ecs"]:
        #     paula_item = Item("Paula")
        # else:
        #     paula_item = Item("Paula")
        # chipset_item.add(paula_item)

    # if c.kickstart_file:
    #     kickstart_item = Item("Custom Kickstart [FIXME]")
    # else:
    #     kickstart_item = Item("Default Kickstart [FIXME]")
    kickstart_item = Item(
        "Kickstart {} Rev {}".format(
            c.int_kickstart_version, c.int_kickstart_revision
        )
    )
    model.add(kickstart_item)

    if c.int_kickstart_ext_sha1 == "5bef3d628ce59cc02a66e6e4ae0da48f60e78f7f":
        kickstart_ext_item = Item("CD32 Extended ROM")
    elif (
        c.int_kickstart_ext_sha1 == "7ba40ffa17e500ed9fed041f3424bd81d9c907be"
    ):
        kickstart_ext_item = Item("CDTV Extended ROM")
    else:
        kickstart_ext_item = InactiveItem("Extended ROM")
    model.add(kickstart_ext_item)

    # if c.uae_cd32cd == "true":
    #     akiko_item = Item("Akiko Chip")
    #     # chipset_item.add(akiko_item)
    #     model.add(akiko_item)

    jit_text = " [JIT]" if c.jit_compiler == "1" else ""
    cpu_item = Item("MC" + c.int_cpu_name + jit_text)
    cpu_item.represents = ["cpu", "uae_cpu_model"]  # FIXME: uae_cpu_type

    if c.int_accelerator_name:
        item = Item("{0}".format(c.int_accelerator_name))
        accelerator_item = model.add(item)
    else:
        accelerator_item = InactiveItem("No Accelerator")
        model.add(accelerator_item)
    accelerator_item.represents = ["accelerator", "uae_cpuboard_type"]

    if accelerator_item.active:
        accelerator_item.add(cpu_item)
    else:
        model.add(cpu_item)

    if accelerator_item.active:
        if c.int_ppc_model:
            item = Item("PowerPC {0}".format(c.int_ppc_model))
            accelerator_item.add(item)

        # model.remove(cpu_item)
        # cpu_item.set_parent(accelerator_item)
        # model.add(cpu_item)

        if c.int_cpuboardmem1_size:
            size = "{0} MB".format(
                int(c.int_cpuboardmem1_size) // (1024 * 1024)
            )
            item = Item("{0} RAM".format(size))
            item.represents = ["uae_cpuboard1mem_size"]
            accelerator_item.add(item)

        item = Item("ROM/Flash [FIXME]")
        item.represents = ["accelerator_rom"]
        accelerator_item.add(item)

        if c.uae_cpuboard_type == "CyberStormPPC":
            # FIXME: other too?
            accelerator_scsi_item = Item("CyberStorm SCSI")
            accelerator_item.add(accelerator_scsi_item)

    if int(c.uae_mmu_model):
        mmu_item = Item("{0} MMU".format(c.uae_mmu_model))
    else:
        mmu_item = InactiveItem("No MMU")
    cpu_item.represents = ["mmu_model"]
    cpu_item.add(mmu_item)

    if int(c.uae_fpu_model):
        fpu_item = Item("{0} FPU".format(c.uae_fpu_model))
    else:
        fpu_item = InactiveItem("No FPU")
        # fpu_item.set_parent(cpu_item)
    fpu_item.represents = ["fpu_model"]
    if c.uae_fpu_model == c.uae_cpu_model:
        cpu_item.add(fpu_item)
    else:
        model.add(fpu_item)

    if int(c.int_chipmem_size) % (1024 * 1024) == 0:
        size = "{0} MB".format(int(c.int_chipmem_size) // (1024 * 1024))
    else:
        size = "{0} KB".format(int(c.int_chipmem_size) // 1024)
    chipmem_item = Item("{0} Chip RAM".format(size))
    chipmem_item.represents = ["chip_memory", "uae_chipmem_size"]
    model.add(chipmem_item)

    if int(c.int_mbresmem_low_size):
        size = "{0} MB".format(int(c.int_mbresmem_low_size) // (1024 * 1024))
        item = Item("{0} Fast RAM".format(size))
    else:
        item = InactiveItem("No Fast RAM")
    item.extra = "Motherboard"
    item.represents = ["uae_a3000mem_size"]
    model.add(item)

    trapdoor_item = ContainerItem("Trapdoor Slot")
    zorro_ii_item = ContainerItem("Zorro II Bus")
    zorro_iii_item = ContainerItem("Zorro III Bus")

    flatten = True

    if int(c.int_bogomem_size):
        if int(c.int_bogomem_size) % (1024 * 1024) == 0:
            size = "{0} MB".format(int(c.int_bogomem_size) // (1024 * 1024))
        else:
            size = "{0} KB".format(int(c.int_bogomem_size) // 1024)
        bogomem_item = Item("{0} Slow RAM".format(size))
    else:
        bogomem_item = InactiveItem("No Slow RAM")
    bogomem_item.represents = ["slow_memory", "uae_bogomem_size"]
    if flatten:
        bogomem_item.extra = "Trapdoor"
        model.add(bogomem_item)
    else:
        trapdoor_item.add(bogomem_item)

    if int(c.int_fastmem_size):
        if int(c.int_fastmem_size) % (1024 * 1024) == 0:
            size = "{0} MB".format(int(c.int_fastmem_size) // (1024 * 1024))
        else:
            size = "{0} KB".format(int(c.int_fastmem_size) // 1024)
        # fastmem_item = Item("{0} Zorro II Fast RAM".format(size))
        fastmem_item = Item("{0} Fast RAM".format(size))
    else:
        fastmem_item = InactiveItem("No Zorro II Fast RAM")
    fastmem_item.represents = ["fast_memory", "uae_fastmem_size"]
    if flatten:
        fastmem_item.extra = "Zorro II"
        model.add(fastmem_item)
    else:
        zorro_ii_item.add(fastmem_item)

    if int(c.int_z3fastmem_size):
        size = "{0} MB".format(int(c.int_z3fastmem_size) // (1024 * 1024))
        # z3fastmem_item = Item("{0} Zorro III Fast RAM".format(size))
        z3fastmem_item = Item("{0} Fast RAM".format(size))
    else:
        z3fastmem_item = InactiveItem("No Zorro III Fast RAM")
    z3fastmem_item.represents = ["zorro_iii_memory", "uae_z3mem_size"]
    if flatten:
        z3fastmem_item.extra = "Zorro III"
        model.add(z3fastmem_item)
    else:
        zorro_iii_item.add(z3fastmem_item)

    if c.uae_rtc != "none":
        rtc_item = Item("{0} RTC".format(c.uae_rtc))
    else:
        rtc_item = InactiveItem("No Real Time Clock")
    rtc_item.represents = ["uae_rtc"]
    if bogomem_item.active and c.uae_rtc == "MSM6242B":
        if flatten:
            # bogomem_item.add(rtc_item)
            model.add(rtc_item)
        else:
            trapdoor_item.add(rtc_item)
    else:
        model.add(rtc_item)

    if c.uae_cd32fmv == "true":
        fmv_item = Item("CD32 FMV Module")
        model.add(fmv_item)

        fmv_rom_item = Item("CD32 FMV ROM")
        fmv_item.add(fmv_rom_item)

    model.add(trapdoor_item)
    model.add(zorro_ii_item)
    model.add(zorro_iii_item)

    # resident_item = ContainerItem("Resident Libraries")
    # model.add(resident_item)
    resident_item = model

    if c.int_uae_boot_rom == "true":
        uae_boot_rom_item = Item("UAE Boot ROM")
        resident_item.add(uae_boot_rom_item)

        uae_resource_item = Item("UAE uae.resource")
        uae_boot_rom_item.add(uae_resource_item)

        if c.uae_bsdsocket_emu == "true":
            bsdsocket_item = Item("UAE bsdsocket.library")
        else:
            bsdsocket_item = InactiveItem("No bsdsocket.library")
        bsdsocket_item.represents = ["bsdsocket_library", "uae_bsdsocket_emu"]
        resident_item.add(bsdsocket_item)

        if c.uae_native_code == "true":
            uaenative_library_item = Item("UAE uaenative.library")
        else:
            uaenative_library_item = InactiveItem("No uaenative.library")
        uaenative_library_item.represents = [
            "uaenative.library",
            "uae_native_code",
        ]
        resident_item.add(uaenative_library_item)

        if c.uae_sana2 == "true":
            uaenet_device_item = Item("UAE uaenet.device")
        else:
            uaenet_device_item = InactiveItem("No uaenet.device")
        uaenet_device_item.represents = ["uae_sana2"]
        resident_item.add(uaenet_device_item)

    # if c.uae_a2065:
    #     a2065_item = Item("A2065 [Network Card]")
    #     a2065_item.represents = ["uae_a2065"]
    #     zorro_ii_item.add(a2065_item)
    #
    #     if c.uae_a2065 == "slirp":
    #         slirp_item = create_slirp_item(c)
    #         a2065_item.add(slirp_item)

    if c.uae_gfxcard_type:
        graphics_card_item = Item(
            "{} {} MB".format(c.int_graphics_card_name, c.uae_gfxcard_size)
        )
        if flatten:
            if c.int_graphics_card_bus == "zorro-ii":
                graphics_card_item.extra = "Zorro II"
            elif c.int_graphics_card_bus == "zorro-iii":
                graphics_card_item.extra = "Zorro III"
            model.add(graphics_card_item)
        elif c.int_graphics_card_bus == "zorro-ii":
            zorro_ii_item.add(graphics_card_item)
        elif c.int_graphics_card_bus == "zorro-iii":
            zorro_iii_item.add(graphics_card_item)
        else:
            model.add(graphics_card_item)

    if c.uae_toccata == "true":
        sound_card_item = Item("Toccata")
        sound_card_item.represents = ["sound_card", "uae_toccata"]
        if flatten:
            sound_card_item.extra = "Zorro II"
            model.add(sound_card_item)
        else:
            zorro_ii_item.add(sound_card_item)

    if c.uae_a2065:
        network_card_item = Item("A2065")
        network_card_item.represents = ["network_card", "uae_a2065"]
        if flatten:
            network_card_item.extra = "Zorro II"
            model.add(network_card_item)
        else:
            zorro_ii_item.add(network_card_item)

    joystick_port_0_item = create_joystick_port_item(c, 0)
    joystick_port_1_item = create_joystick_port_item(c, 1)
    joystick_port_2_item = create_joystick_port_item(c, 2)
    joystick_port_3_item = create_joystick_port_item(c, 3)

    model.add(joystick_port_0_item)
    model.add(joystick_port_1_item)
    # parallel_port_item = Item("Parallel Port")

    if joystick_port_2_item.active or joystick_port_3_item.active:
        parallel_port_item = Item("Parallel Port Joystick Adapter")
        model.add(parallel_port_item)
        parallel_port_item.add(joystick_port_2_item)
        parallel_port_item.add(joystick_port_3_item)
    else:
        parallel_port_item = InactiveItem("No Parallel Port Device")
        model.add(parallel_port_item)

    for i in range(4):
        drive_type = getattr(c, "uae_floppy{0}type".format(i))
        if drive_type == "0":
            description = '3.5" DD'
        elif drive_type == "1":
            description = '3.5" HD'
        elif drive_type == "2":
            description = '5.25" SD'
        elif drive_type == "3":
            description = '3.5" DD (ESCOM)'
        else:
            description = None
        if description:
            drive_item = Item(
                "{0} Floppy Drive [DF{1}]".format(description, i)
            )
        else:
            drive_item = InactiveItem("No Floppy Drive [DF{0}]".format(i))
        drive_item.represents = [
            "floppy_drive_{0}".format(i),
            "uae_df{0}type".format(i),
        ]
        model.add(drive_item)
        # path = getattr(c, "uae_floppy{0}".format(i))
        # if path:
        #     name = os.path.basename(path)
        #     disk_item = Item(name)
        #     drive_item.add(disk_item)

    return model



import fsui
from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs
from launcher.i18n import gettext
from launcher.launcher_signal import LauncherSignal
from launcher.scanner import Scanner
from launcher.settings.scan_paths_group import ScanPathsGroup

TIMER_INTERVAL = 100


class ScanDialog(fsui.Window):
    @classmethod
    def refresh_game_database(cls, window):
        return cls(
            window, minimal=True, interactive=False, scan_for_files=False
        )

    def __init__(
        self, parent, minimal=False, interactive=True, scan_for_files=True
    ):
        super().__init__(parent, gettext("Update File Database"))
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        buttons.create_close_button()

        self.layout.add_spacer(640, 0)

        self.interactive = interactive
        self.scan_for_files = scan_for_files
        self.update_game_database = False
        # self.update_game_database = True

        if not minimal:
            self.scan_kickstart_group = ScanKickstartGroup(self)
            layout.add(self.scan_kickstart_group, fill=True)

            layout.add_spacer(20)

            label = fsui.HeadingLabel(
                self, gettext("Scan for Kickstarts, Files and Configurations")
            )
            layout.add(label, margin_bottom=10)

            self.scan_paths_group = ScanPathsGroup(self)
            layout.add(self.scan_paths_group, fill=True, margin=0)

            layout.add_spacer(20)

        self.scan_progress_group = ScanProgressGroup(self)
        layout.add(self.scan_progress_group, fill=True)

        # layout.add_spacer(20)
        # layout.add_spacer(20)

        # hor_layout = fsui.HorizontalLayout()
        # layout.add(hor_layout, fill=True)

        # hor_layout.add_spacer(10, expand=True)
        if interactive:
            self.scan_button = buttons.add_button(
                fsui.Button(buttons, gettext("Scan"))
            )
            # self.scan_button = fsui.Button(self, _("Scan"))
            # hor_layout.add(self.scan_button)
            # hor_layout.add_spacer(10)
            self.scan_button.activated.connect(self.on_scan_button)
        else:
            self.scan_button = None

        # self.stop_button = fsui.Button(self, _("Abort"))
        # hor_layout.add(self.stop_button)
        self.stop_button = buttons.add_button(
            fsui.Button(buttons, gettext("Stop"))
        )
        self.stop_button.activated.connect(self.on_stop_button)

        # layout.add_spacer(10)

        # self.set_size(layout.get_min_size())
        # self.center_on_parent()

        self.old_title = ""
        self.old_status = ""
        self.has_started_scan = False

        self.timer = fsui.IntervalTimer(TIMER_INTERVAL)
        self.timer.activated.connect(self.on_timer)
        self.closed.connect(self.timer.stop)

        if not self.interactive:
            self.start_scan()

    def on_close(self):
        Scanner.stop_flag = True

    def set_scan_title(self, text):
        if not text:
            return
        if text == self.old_title:
            return
        self.old_title = text
        self.scan_progress_group.title_label.set_text(text)

    def set_scan_status(self, text):
        if not text:
            return
        if text == self.old_status:
            return
        self.old_status = text
        self.scan_progress_group.status_label.set_text(text)

    def on_timer(self):
        if not Scanner.running:
            if self.has_started_scan:
                if Scanner.error:
                    self.set_scan_title(gettext("Scan error"))
                    self.set_scan_status(Scanner.error)
                else:
                    if not self.interactive:
                        self.end_modal(True)
                        return
                    self.set_scan_title(gettext("Scan complete"))
                    self.set_scan_status(
                        gettext("Click 'Scan' button if you want to re-scan")
                    )
            else:
                self.set_scan_title(gettext("No scan in progress"))
                self.set_scan_status(
                    gettext("Click 'Scan' button to start scan")
                )
            if self.scan_button is not None:
                self.scan_button.enable()
            self.stop_button.disable()
            # self.close_button.enable()
            return

        status = Scanner.status
        self.set_scan_title(status[0])
        self.set_scan_status(status[1])

    def on_scan_button(self):
        self.start_scan()

    def start_scan(self):
        if self.scan_button is not None:
            self.scan_button.disable()
        self.has_started_scan = True
        self.set_scan_title(gettext("Starting scan"))
        self.set_scan_status(gettext("Please wait..."))
        paths = ScanPathsGroup.get_search_path()

        # self.close_button.disable()
        self.stop_button.enable()

        Scanner.start(
            paths,
            scan_for_files=self.scan_for_files,
            update_game_database=self.update_game_database,
            purge_other_dirs=True,
        )

    # def on_close_button(self):
    #     self.end_modal(False)

    # noinspection PyMethodMayBeStatic
    def on_stop_button(self):
        Scanner.stop_flag = True
        # self.close_button.enable()


class KickstartStatusGroup(fsui.Group):
    def __init__(self, parent, title, model):
        self.model = model
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        self.ok_image = fsui.Image("launcher:res/ok_emblem.png")
        self.na_image = fsui.Image("launcher:res/na_emblem.png")

        self.icon = fsui.ImageView(self, self.na_image)
        self.layout.add(self.icon)

        self.layout.add_spacer(10)
        self.label = fsui.Label(self, title)
        self.layout.add(self.label)
        self.update()
        LauncherSignal.add_listener("scan_done", self)

    def on_destroy(self):
        LauncherSignal.remove_listener("scan_done", self)

    def on_scan_done_signal(self):
        self.update()

    def update(self):
        amiga = Amiga.get_model_config(self.model)
        for sha1 in amiga["kickstarts"]:
            if fsgs.file.find_by_sha1(sha1):
                self.icon.set_image(self.ok_image)
                return
        self.icon.set_image(self.na_image)


class ScanKickstartGroup(fsui.Group):
    def __init__(self, parent):
        fsui.Group.__init__(self, parent)

        self.layout = fsui.VerticalLayout()
        label = fsui.HeadingLabel(
            self, gettext("Available Kickstart Versions")
        )
        self.layout.add(label, margin_bottom=10)

        icon_layout = fsui.HorizontalLayout()
        self.layout.add(icon_layout, fill=True)

        icon_layout.add_spacer(20)
        image = fsui.Image("launcher:res/kickstart.png")
        self.image_view = fsui.ImageView(self, image)
        icon_layout.add(self.image_view, valign=0.0, margin_right=10)

        vert_layout = fsui.VerticalLayout()
        icon_layout.add(vert_layout, fill=True, expand=True)

        vert_layout.add_spacer(0)

        label = fsui.Label(
            self,
            gettext(
                "You should have kickstart files for "
                "each Amiga model you want to use:"
            ),
        )
        vert_layout.add(label, margin_bottom=0)

        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True)

        self.kickstart_groups = []

        column_layout = fsui.VerticalLayout()
        hori_layout.add(column_layout, expand=True, fill=True, margin=10)

        self.add_kickstart_group(column_layout, "Amiga 1000", "A1000")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 500", "A500")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 500+", "A500+")

        column_layout = fsui.VerticalLayout()
        hori_layout.add(column_layout, expand=True, fill=True, margin=10)

        self.add_kickstart_group(column_layout, "Amiga 600", "A600")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 1200", "A1200")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 3000", "A3000")

        column_layout = fsui.VerticalLayout()
        hori_layout.add(column_layout, expand=True, fill=True, margin=10)

        self.add_kickstart_group(column_layout, "Amiga 4000", "A4000/040")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga CD32", "CD32")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Commodore CDTV", "CDTV")

    def add_kickstart_group(self, layout, title, model):
        group = KickstartStatusGroup(self, title, model)
        self.kickstart_groups.append(group)
        layout.add(group, fill=True)


class ScanProgressGroup(fsui.Group):
    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()
        # self.layout.padding_left = 10
        # self.layout.padding_top = 10
        # self.layout.padding_right = 10
        # self.layout.padding_bottom = 10

        # #image = fsui.Image("launcher:res/search_group.png")
        # #self.image_view = fsui.ImageView(self, image)
        # self.layout.add_spacer(20)
        # #self.layout.add(self.image_view, valign=0.0)
        # self.layout.add_spacer(48)
        # self.layout.add_spacer(20)

        self.layout2 = fsui.VerticalLayout()
        self.layout.add(self.layout2, fill=True, expand=True)

        self.title_label = fsui.HeadingLabel(self, "")
        self.layout2.add(self.title_label, fill=True)

        # self.layout2.add_spacer(10)
        # hor_layout = fsui.HorizontalLayout()
        # self.layout2.add(hor_layout)

        # self.scan_label = fsui.Label(self, _("Scan for:"))
        # hor_layout.add(self.scan_label)
        # hor_layout.add_spacer(10)

        # self.scan_roms = fsui.CheckBox(self, _("ROMs"))
        # if Settings.get("scan_roms") == "1":
        #     self.scan_roms.check()
        # self.scan_roms.on_changed = self.on_change
        # hor_layout.add(self.scan_roms)
        # hor_layout.add_spacer(10)

        # self.scan_files = fsui.CheckBox(self, _("Game Files"))
        # if Settings.get("scan_files") == "1":
        #     self.scan_files.check()
        # self.scan_files.on_changed = self.on_change
        # hor_layout.add(self.scan_files)
        # hor_layout.add_spacer(10)

        # self.scan_configs = fsui.CheckBox(self, _("Configurations"))
        # if Settings.get("scan_configs") == "1":
        #     self.scan_configs.check()
        # self.scan_configs.on_changed = self.on_change
        # hor_layout.add(self.scan_configs)
        # hor_layout.add_spacer(10)

        self.layout2.add_spacer(10)
        self.status_label = fsui.Label(self, "")
        self.layout2.add(self.status_label, fill=True)

        # def on_change(self):
        #     value = "1" if self.scan_roms.is_checked() else "0"
        #     Settings.set("scan_roms", value)
        #     value = "1" if self.scan_files.is_checked() else "0"
        #     Settings.set("scan_files", value)
        #     value = "1" if self.scan_configs.is_checked() else "0"
        #     Settings.set("scan_configs", value)



from arcade.glui.opengl import fs_emu_blending
from OpenGL import GL as gl

from arcade.glui.render import Render

from .state import State


class Dialog(object):
    @classmethod
    def get_current(cls):
        return State.get().dialog

    def __init__(self):
        self.width = 1.6
        self.height = 1.0
        self.background_texture = None
        self.background_texture_offset = [0, 0]
        self.background_color = (0.4, 0.4, 0.4, 1.0)
        # make sure that the dialog is rendered at least once
        # after being created
        Render.get().dirty = True

    def show(self):
        State.get().dialog = self
        State.get().dialog_time = State.get().time

    def render(self):
        Render.get().ortho_perspective()
        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_blending(True)
        gl.glPushMatrix()
        gl.glTranslatef(-self.width / 2, -self.height / 2, 0.0)
        self.render_background()
        self.render_content()
        gl.glPopMatrix()
        gl.glEnable(gl.GL_DEPTH_TEST)
        pass

    def render_background(self):
        if self.background_texture:
            gl.glBindTexture(gl.GL_TEXTURE_2D, self.background_texture)
            gl.glColor3f(1.0, 1.0, 1.0)
        else:
            gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
            # r, g, b a = self.background_color
            gl.glColor4f(*self.background_color)
        gl.glBegin(gl.GL_QUADS)
        bgx, bgy = self.background_texture_offset
        gl.glTexCoord(bgx + 0.0, bgy + 1.0)
        gl.glVertex2f(0, 0)
        gl.glTexCoord(bgx + 1.0, bgy + 1.0)
        gl.glVertex2f(self.width, 0)
        gl.glTexCoord(bgx + 1.0, bgy + 0.0)
        gl.glVertex2f(self.width, self.height)
        gl.glTexCoord(bgx + 0.0, bgy + 0.0)
        gl.glVertex2f(0, self.height)
        gl.glEnd()

    def render_content(self):
        pass

    def destroy(self):
        pass

    def close(self):
        if self == State.get().dialog:
            self.destroy()
            State.get().dialog = None
        else:
            print("WARNING: close called on non-active dialog")



import os

from fsgs.drivers.dolphindriver import DolphinDriver, DolphinInputMapper
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

NGC_PLATFORM_NAME = "GameCube"
NGC_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "ngc",
}


class GameCubePlatform(Platform):
    # FIXME: Move to init instead
    PLATFORM_NAME = NGC_PLATFORM_NAME

    def driver(self, fsgc):
        return GameCubeDriver(fsgc)

    def loader(self, fsgc):
        return GameCubeLoader(fsgc)


class GameCubeLoader(SimpleLoader):
    pass


class GameCubeDriver(DolphinDriver):
    PORTS = [
        {"description": "Controller 1", "types": [NGC_CONTROLLER]},
        {"description": "Controller 2", "types": [NGC_CONTROLLER]},
        {"description": "Controller 3", "types": [NGC_CONTROLLER]},
        {"description": "Controller 4", "types": [NGC_CONTROLLER]},
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.helper = GameCubeHelper(self.options)

    def dolphin_configure_core(self, f):
        # find devices now (need to know how many devices to
        # specify in dolphin.ini)
        # devices = self.context.input.get_devices(1, 4)
        # configure dolphin.ini
        device_count = 0
        for i, port in enumerate(self.ports):
            if port.device:
                device_count += 1
        for i in range(device_count):
            f.write("SIDevice{0} = 150994944\n".format(i))

        # FIXME:
        memcard_region = "USA"

        f.write(
            "MemcardA = {dir}/MemoryCardA.{region}.raw\n".format(
                dir=self.get_state_dir(), region=memcard_region
            )
        )
        f.write(
            "MemcardB = {dir}/MemoryCardB.{region}.raw\n".format(
                dir=self.get_state_dir(), region=memcard_region
            )
        )

    def dolphin_configure_input(self):
        # devices = self.context.input.get_devices(1, 4)
        input_mapping = {
            "A": "Buttons/A",
            "B": "Buttons/B",
            "X": "Buttons/X",
            "Y": "Buttons/Y",
            "Z": "Buttons/Z",
            "START": "Buttons/Start",
            "STICK_UP": "Main Stick/Up",
            "STICK_DOWN": "Main Stick/Down",
            "STICK_LEFT": "Main Stick/Left",
            "STICK_RIGHT": "Main Stick/Right",
            "C_UP": "C-Stick/Up",  # or z?
            "C_DOWN": "C-Stick/Down",
            "C_LEFT": "C-Stick/Left",
            "C_RIGHT": "C-Stick/Right",
            "L": "Triggers/L",
            "R": "Triggers/R",
            "DPAD_UP": "D-Pad/Up",
            "DPAD_DOWN": "D-Pad/Down",
            "DPAD_LEFT": "D-Pad/Left",
            "DPAD_RIGHT": "D-Pad/Right",
            "L_ANALOG": "Triggers/L-Analog",
            "R_ANALOG": "Triggers/R-Analog",
            # FIXME: ADD MODIFIER?
        }
        temp_dir = self.temp_dir("dolphin")
        input_config_file = os.path.join(
            temp_dir.path, "user", "Config", "GCPadNew.ini"
        )
        f = open(input_config_file, "w")
        for i, port in enumerate(self.ports):
            if not port.device:
                continue
            f.write("[GCPad{num}]\n".format(num=i + 1))
            if port.device.is_keyboard():
                f.write("Device = DInput/0/Keyboard Mouse\n")
            else:
                type_index = int(port.device.id.rsplit("#", 1)[1]) - 1
                f.write(
                    "Device = SDL/{index}/{name}\n".format(
                        index=type_index, name=port.device.sdl_name
                    )
                )
            mapper = DolphinInputMapper(port, input_mapping)
            for key, value in mapper.items():
                if isinstance(key, tuple):
                    for k in key:
                        f.write("{key} = {value}\n".format(key=k, value=value))
                else:
                    f.write("{key} = {value}\n".format(key=key, value=value))


class GameCubeHelper:
    def __init__(self, options):
        self.options = options



# Copyright (c) 2008 Duncan Fordyce
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import inspect
import logging
import traceback

from oyoyo import helpers
from oyoyo.parse import parse_nick


def protected(func):
    """ decorator to protect functions from being called """
    func.protected = True
    return func


class CommandError(Exception):
    def __init__(self, cmd):
        self.cmd = cmd


class NoSuchCommandError(CommandError):
    def __str__(self):
        return 'No such command "%s"' % ".".join(self.cmd)


class ProtectedCommandError(CommandError):
    def __str__(self):
        return 'Command "%s" is protected' % ".".join(self.cmd)


class CommandHandler(object):
    """ The most basic CommandHandler """

    def __init__(self, client):
        self.client = client

    @protected
    def get(self, in_command_parts):
        """ finds a command
        commands may be dotted. each command part is checked that it does
        not start with and underscore and does not have an attribute
        "protected". if either of these is true, ProtectedCommandError
        is raised.
        its possible to pass both "command.sub.func" and
        ["command", "sub", "func"].
        """
        if isinstance(in_command_parts, (str, bytes)):
            in_command_parts = in_command_parts.split(".")
        command_parts = in_command_parts[:]

        p = self
        while command_parts:
            cmd = command_parts.pop(0)
            if cmd.startswith('_'):
                raise ProtectedCommandError(in_command_parts)

            try:
                f = getattr(p, cmd)
            except AttributeError:
                raise NoSuchCommandError(in_command_parts)

            if hasattr(f, 'protected'):
                raise ProtectedCommandError(in_command_parts)

            if isinstance(f, CommandHandler) and command_parts:
                return f.get(command_parts)
            p = f

        return f

    @protected
    def run(self, command, *args):
        """ finds and runs a command """
        logging.debug("processCommand %s(%s)" % (command, args))
        # print(command, args)

        try:
            f = self.get(command)
        except NoSuchCommandError:
            self.__unhandled__(command, *args)
            return

        logging.debug('f %s' % f)

        try:
            f(*args)
        except Exception as e:
            logging.error('command raised %s' % e)
            logging.error(traceback.format_exc())
            raise CommandError(command)

    @protected
    def __unhandled__(self, cmd, *args):
        """The default handler for commands. Override this method to
        apply custom behavior (example, printing) unhandled commands.
        """
        logging.debug('unhandled command %s(%s)' % (cmd, args))


class DefaultCommandHandler(CommandHandler):
    """ CommandHandler that provides methods for the normal operation of IRC.
    If you want your bot to properly respond to pings, etc, you should subclass this.
    """

    def ping(self, prefix, server):
        self.client.send('PONG', server)


# noinspection PyPep8Naming
class DefaultBotCommandHandler(CommandHandler):
    """ default command handler for bots. methods/attributes are made
    available as commands """

    @protected
    def getVisibleCommands(self, obj=None):
        test = (lambda x: isinstance(x, CommandHandler) or
                inspect.ismethod(x) or inspect.isfunction(x))
        members = inspect.getmembers(obj or self, test)
        return [m for m, _ in members
            if (not m.startswith('_') and
                not hasattr(getattr(obj, m), 'protected'))]

    def help(self, sender, dest, arg=None):
        """list all available commands or get help on a specific command"""
        logging.info('help sender=%s dest=%s arg=%s' % (sender, dest, arg))
        if not arg:
            commands = self.getVisibleCommands()
            commands.sort()
            helpers.msg(self.client, dest,
                        "available commands: %s" % " ".join(commands))
        else:
            try:
                f = self.get(arg)
            except CommandError as e:
                helpers.msg(self.client, dest, str(e))
                return

            doc = f.__doc__.strip() if f.__doc__ else "No help available"

            if not inspect.ismethod(f):
                sub_commands = self.getVisibleCommands(f)
                if sub_commands:
                    doc += " [sub commands: %s]" % " ".join(sub_commands)

            helpers.msg(self.client, dest, "%s: %s" % (arg, doc))


# noinspection PyPep8Naming
class BotCommandHandler(DefaultCommandHandler):
    """ complete command handler for bots """

    def __init__(self, client, command_handler):
        DefaultCommandHandler.__init__(self, client)
        self.command_handler = command_handler

    def privmsg(self, prefix, dest, msg):
        self.tryBotCommand(prefix, dest, msg)

    @protected
    def tryBotCommand(self, prefix, dest, msg):
        """ tests a command to see if its a command for the bot, returns True
        and calls self.processBotCommand(cmd, sender) if its is.
        """

        logging.debug("tryBotCommand('%s' '%s' '%s')" % (prefix, dest, msg))

        if dest == self.client.nick:
            dest = parse_nick(prefix)[0]
        elif msg.startswith(self.client.nick):
            msg = msg[len(self.client.nick)+1:]
        else:
            return False

        msg = msg.strip()

        parts = msg.split(' ', 1)
        command = parts[0]
        arg = parts[1:]

        try:
            self.command_handler.run(command, prefix, dest, *arg)
        except CommandError as e:
            helpers.msg(self.client, dest, str(e))
        return True



from fsgs.util import sdl2

# with open(os.expanduser('~/Desktop/keys.txt'), 'wb') as f:
#     for key in sorted(sdl_key_codes.keys()):
#         value = str(sdl_key_codes[key])
#         sdl_name = 'SDLK_' + key.upper()
#         try:
#             dik_name = 'DIK_' + sdlk_to_dik[key.upper()]
#         except KeyError:
#             dik_name = ''
#         try:
#             dik_code = str(dinput_key_codes[key.upper()])
#         except KeyError:
#             dik_code = ''
#         f.write("    '" + sdl_name + "': (" + value + ", '" + dik_name +
#                 "', " + dik_code + '),\n')


class Key(object):
    def __init__(self, name):
        self.name = name

    @property
    def sdl_name(self):
        return self.name

    # Deprecated
    @property
    def sdl_code(self):
        return key_table[self.name][0]

    @property
    def sdl2_key_code(self):
        return key_table[self.name][0]

    @property
    def dinput_name(self):
        return key_table[self.name][1]

    @property
    def dinput_code(self):
        return key_table[self.name][2]

    @property
    def sdl2_scan_code(self):
        return key_table[self.name][3]

    def __str__(self):
        return "<Key {}>".format(self.name)


class Keyboard(object):
    @staticmethod
    def key(name):
        # try:
        #     code = name.key.keysym.sym
        # except AttributeError:
        #     pass
        print("key({})".format(name))
        try:
            code = name["key"]
        except (KeyError, TypeError):
            pass
        else:
            name = sdl_key_code_to_name[code]
        if isinstance(name, int):
            name = sdl_key_code_to_name[name]
        name = name.upper()
        if name.startswith("SDLK_"):
            pass
        else:
            name = "SDLK_" + name
        return Key(name)


# class Keyboard(object):
#     @staticmethod
#     def key(name):
#         if isinstance(name, int):
#             name = sdl_key_code_to_name[name]
#         name = name.upper()
#         if name.startswith("SDLK_"):
#             pass
#         else:
#             name = "SDLK_" + name
#         return Key(name)


key_table = {
    "SDLK_NO_KEY": (0, "DIK_NO_KEY", 0, 0),
    "SDLK_0": (48, "DIK_0", 11, sdl2.SDL_SCANCODE_0),
    "SDLK_1": (49, "DIK_1", 2, sdl2.SDL_SCANCODE_1),
    "SDLK_2": (50, "DIK_2", 3, sdl2.SDL_SCANCODE_2),
    "SDLK_3": (51, "DIK_3", 4, sdl2.SDL_SCANCODE_3),
    "SDLK_4": (52, "DIK_4", 5, sdl2.SDL_SCANCODE_4),
    "SDLK_5": (53, "DIK_5", 6, sdl2.SDL_SCANCODE_5),
    "SDLK_6": (54, "DIK_6", 7, sdl2.SDL_SCANCODE_6),
    "SDLK_7": (55, "DIK_7", 8, sdl2.SDL_SCANCODE_7),
    "SDLK_8": (56, "DIK_8", 9, sdl2.SDL_SCANCODE_8),
    "SDLK_9": (57, "DIK_9", 10, sdl2.SDL_SCANCODE_9),
    "SDLK_A": (97, "DIK_A", 30, sdl2.SDL_SCANCODE_A),
    "SDLK_AMPERSAND": (38, "", 0),
    "SDLK_ASTERISK": (42, "", 0),
    "SDLK_AT": (64, "DIK_AT", 145),
    "SDLK_B": (98, "DIK_B", 48),
    "SDLK_BACKQUOTE": (96, "DIK_GRAVE", 0),
    "SDLK_BACKSLASH": (92, "DIK_BACKSLASH", 43),
    "SDLK_BACKSPACE": (sdl2.SDLK_BACKSPACE, "DIK_BACK", 14),
    "SDLK_BREAK": (318, "", 0),
    "SDLK_C": (99, "DIK_C", 46, sdl2.SDL_SCANCODE_C),
    "SDLK_CAPSLOCK": (301, "DIK_CAPITAL", 58),
    "SDLK_CARET": (94, "", 0),
    "SDLK_CLEAR": (12, "", 0),
    "SDLK_COLON": (58, "DIK_COLON", 146),
    "SDLK_COMMA": (44, "DIK_COMMA", 51),
    "SDLK_COMPOSE": (314, "", 0),
    "SDLK_D": (100, "DIK_D", 32, sdl2.SDL_SCANCODE_D),
    "SDLK_DELETE": (sdl2.SDLK_DELETE, "DIK_DELETE", 211),
    "SDLK_DOLLAR": (36, "", 0),
    "SDLK_DOWN": (sdl2.SDLK_DOWN, "DIK_DOWN", 208, sdl2.SDL_SCANCODE_DOWN),
    "SDLK_E": (101, "DIK_E", 18),
    "SDLK_END": (sdl2.SDLK_END, "DIK_END", 207),
    "SDLK_EQUALS": (61, "DIK_EQUALS", 13),
    "SDLK_ESCAPE": (sdl2.SDLK_ESCAPE, "DIK_ESCAPE", 1),
    "SDLK_EURO": (321, "", 0),
    "SDLK_EXCLAIM": (33, "", 0),
    "SDLK_F": (102, "DIK_F", 33),
    "SDLK_F1": (282, "DIK_F1", 59),
    "SDLK_F10": (291, "DIK_F10", 68),
    "SDLK_F11": (292, "DIK_F11", 87),
    "SDLK_F12": (293, "DIK_F12", 88),
    "SDLK_F13": (294, "DIK_F13", 100),
    "SDLK_F14": (295, "DIK_F14", 101),
    "SDLK_F15": (296, "DIK_F15", 102),
    "SDLK_F2": (283, "DIK_F2", 60),
    "SDLK_F3": (284, "DIK_F3", 61),
    "SDLK_F4": (285, "DIK_F4", 62),
    "SDLK_F5": (286, "DIK_F5", 63),
    "SDLK_F6": (287, "DIK_F6", 64),
    "SDLK_F7": (288, "DIK_F7", 65),
    "SDLK_F8": (289, "DIK_F8", 66),
    "SDLK_F9": (290, "DIK_F9", 67),
    "SDLK_G": (103, "DIK_G", 34),
    "SDLK_GREATER": (62, "", 0),
    "SDLK_H": (104, "DIK_H", 35),
    "SDLK_HASH": (35, "", 0),
    "SDLK_HELP": (315, "", 0),
    "SDLK_HOME": (sdl2.SDLK_HOME, "DIK_HOME", 199),
    "SDLK_I": (105, "DIK_I", 23),
    "SDLK_INSERT": (sdl2.SDLK_INSERT, "DIK_INSERT", 210),
    "SDLK_J": (106, "DIK_J", 36),
    "SDLK_K": (107, "DIK_K", 37),
    "SDLK_KP0": (sdl2.SDLK_KP_0, "DIK_NUMPAD0", 82),
    "SDLK_KP1": (sdl2.SDLK_KP_1, "DIK_NUMPAD1", 79),
    "SDLK_KP2": (sdl2.SDLK_KP_2, "DIK_NUMPAD2", 80),
    "SDLK_KP3": (sdl2.SDLK_KP_3, "DIK_NUMPAD3", 81),
    "SDLK_KP4": (sdl2.SDLK_KP_4, "DIK_NUMPAD4", 75),
    "SDLK_KP5": (sdl2.SDLK_KP_5, "DIK_NUMPAD5", 76),
    "SDLK_KP6": (sdl2.SDLK_KP_6, "DIK_NUMPAD6", 77),
    "SDLK_KP7": (sdl2.SDLK_KP_7, "DIK_NUMPAD7", 71),
    "SDLK_KP8": (sdl2.SDLK_KP_8, "DIK_NUMPAD8", 72),
    "SDLK_KP9": (sdl2.SDLK_KP_9, "DIK_NUMPAD9", 73),
    "SDLK_KP_DIVIDE": (267, "DIK_DIVIDE", 181),
    "SDLK_KP_ENTER": (271, "DIK_NUMPADENTER", 156),
    "SDLK_KP_EQUALS": (272, "", 141),
    "SDLK_KP_MINUS": (269, "DIK_SUBTRACT", 74),
    "SDLK_KP_MULTIPLY": (268, "DIK_MULTIPLY", 55),
    "SDLK_KP_PERIOD": (sdl2.SDLK_KP_PERIOD, "DIK_DECIMAL", 83),
    "SDLK_KP_PLUS": (270, "DIK_ADD", 78),
    "SDLK_L": (108, "DIK_L", 38, sdl2.SDL_SCANCODE_L),
    "SDLK_LALT": (308, "DIK_LMENU", 56, sdl2.SDL_SCANCODE_LALT),
    "SDLK_LCTRL": (
        sdl2.SDLK_LCTRL,
        "DIK_LCONTROL",
        29,
        sdl2.SDL_SCANCODE_LCTRL,
    ),
    "SDLK_LEFT": (sdl2.SDLK_LEFT, "DIK_LEFT", 203, sdl2.SDL_SCANCODE_LEFT),
    "SDLK_LEFTBRACKET": (91, "DIK_LBRACKET", 26),
    "SDLK_LEFTPAREN": (40, "", 0),
    "SDLK_LESS": (60, "", 0),
    "SDLK_LMETA": (310, "", 0),
    "SDLK_LSHIFT": (
        sdl2.SDLK_LSHIFT,
        "DIK_LSHIFT",
        42,
        sdl2.SDL_SCANCODE_LSHIFT,
    ),
    "SDLK_LSUPER": (311, "DIK_LWIN", 219),  # SDL_SCANCODE_LGUI ?
    "SDLK_M": (109, "DIK_M", 50),
    "SDLK_MENU": (319, "", 0),
    "SDLK_MINUS": (45, "DIK_MINUS", 12),
    "SDLK_MODE": (313, "", 0),
    "SDLK_N": (110, "DIK_N", 49),
    "SDLK_NUMLOCK": (300, "DIK_NUMLOCK", 69),
    "SDLK_O": (111, "DIK_O", 24),
    "SDLK_P": (112, "DIK_P", 25),
    "SDLK_PAGEDOWN": (sdl2.SDLK_PAGEDOWN, "DIK_NEXT", 209),
    "SDLK_PAGEUP": (sdl2.SDLK_PAGEUP, "DIK_PRIOR", 201),
    "SDLK_PAUSE": (19, "", 0),
    "SDLK_PERIOD": (46, "DIK_PERIOD", 52),
    "SDLK_PLUS": (43, "", 0),
    "SDLK_POWER": (320, "", 0),
    "SDLK_PRINT": (316, "", 0),
    "SDLK_Q": (113, "DIK_Q", 16),
    "SDLK_QUESTION": (63, "", 0),
    "SDLK_QUOTE": (39, "DIK_APOSTROPHE", 40),
    "SDLK_QUOTEDBL": (34, "", 0),
    "SDLK_R": (114, "DIK_R", 19),
    "SDLK_RALT": (307, "DIK_RMENU", 184),
    "SDLK_RCTRL": (
        sdl2.SDLK_RCTRL,
        "DIK_RCONTROL",
        157,
        sdl2.SDL_SCANCODE_RCTRL,
    ),
    "SDLK_RETURN": (
        sdl2.SDLK_RETURN,
        "DIK_RETURN",
        28,
        sdl2.SDL_SCANCODE_RETURN,
    ),
    "SDLK_RIGHT": (sdl2.SDLK_RIGHT, "DIK_RIGHT", 205, sdl2.SDL_SCANCODE_RIGHT),
    "SDLK_RIGHTBRACKET": (93, "DIK_RBRACKET", 27),
    "SDLK_RIGHTPAREN": (41, "", 0),
    "SDLK_RMETA": (309, "", 0),
    "SDLK_RSHIFT": (
        sdl2.SDLK_RSHIFT,
        "DIK_RSHIFT",
        54,
        sdl2.SDL_SCANCODE_RSHIFT,
    ),
    "SDLK_RSUPER": (312, "DIK_RWIN", 220),
    "SDLK_S": (115, "DIK_S", 31, sdl2.SDL_SCANCODE_S),
    "SDLK_SCROLLOCK": (302, "DIK_SCROLL", 70),
    "SDLK_SEMICOLON": (59, "DIK_SEMICOLON", 39),
    "SDLK_SLASH": (47, "DIK_SLASH", 53),
    "SDLK_SPACE": (sdl2.SDLK_SPACE, "DIK_SPACE", 57, sdl2.SDL_SCANCODE_SPACE),
    "SDLK_SYSREQ": (317, "DIK_SYSRQ", 183),
    "SDLK_T": (116, "DIK_T", 20),
    "SDLK_TAB": (9, "DIK_TAB", 15),
    "SDLK_U": (117, "DIK_U", 22),
    "SDLK_UNDERSCORE": (95, "", 147),
    "SDLK_UNDO": (322, "", 0),
    "SDLK_UP": (sdl2.SDLK_UP, "DIK_UP", 200, sdl2.SDL_SCANCODE_UP),
    "SDLK_V": (118, "DIK_V", 47),
    "SDLK_W": (119, "DIK_W", 17),
    "SDLK_WORLD_0": (160, ""),
    "SDLK_WORLD_1": (161, ""),
    "SDLK_WORLD_10": (170, ""),
    "SDLK_WORLD_11": (171, ""),
    "SDLK_WORLD_12": (172, ""),
    "SDLK_WORLD_13": (173, ""),
    "SDLK_WORLD_14": (174, ""),
    "SDLK_WORLD_15": (175, ""),
    "SDLK_WORLD_16": (176, ""),
    "SDLK_WORLD_17": (177, ""),
    "SDLK_WORLD_18": (178, ""),
    "SDLK_WORLD_19": (179, ""),
    "SDLK_WORLD_2": (162, ""),
    "SDLK_WORLD_20": (180, ""),
    "SDLK_WORLD_21": (181, ""),
    "SDLK_WORLD_22": (182, ""),
    "SDLK_WORLD_23": (183, ""),
    "SDLK_WORLD_24": (184, ""),
    "SDLK_WORLD_25": (185, ""),
    "SDLK_WORLD_26": (186, ""),
    "SDLK_WORLD_27": (187, ""),
    "SDLK_WORLD_28": (188, ""),
    "SDLK_WORLD_29": (189, ""),
    "SDLK_WORLD_3": (163, ""),
    "SDLK_WORLD_30": (190, ""),
    "SDLK_WORLD_31": (191, ""),
    "SDLK_WORLD_32": (192, ""),
    "SDLK_WORLD_33": (193, ""),
    "SDLK_WORLD_34": (194, ""),
    "SDLK_WORLD_35": (195, ""),
    "SDLK_WORLD_36": (196, ""),
    "SDLK_WORLD_37": (197, ""),
    "SDLK_WORLD_38": (198, ""),
    "SDLK_WORLD_39": (199, ""),
    "SDLK_WORLD_4": (164, ""),
    "SDLK_WORLD_40": (200, ""),
    "SDLK_WORLD_41": (201, ""),
    "SDLK_WORLD_42": (202, ""),
    "SDLK_WORLD_43": (203, ""),
    "SDLK_WORLD_44": (204, ""),
    "SDLK_WORLD_45": (205, ""),
    "SDLK_WORLD_46": (206, ""),
    "SDLK_WORLD_47": (207, ""),
    "SDLK_WORLD_48": (208, ""),
    "SDLK_WORLD_49": (209, ""),
    "SDLK_WORLD_5": (165, ""),
    "SDLK_WORLD_50": (210, ""),
    "SDLK_WORLD_51": (211, ""),
    "SDLK_WORLD_52": (212, ""),
    "SDLK_WORLD_53": (213, ""),
    "SDLK_WORLD_54": (214, ""),
    "SDLK_WORLD_55": (215, ""),
    "SDLK_WORLD_56": (216, ""),
    "SDLK_WORLD_57": (217, ""),
    "SDLK_WORLD_58": (218, ""),
    "SDLK_WORLD_59": (219, ""),
    "SDLK_WORLD_6": (166, ""),
    "SDLK_WORLD_60": (220, ""),
    "SDLK_WORLD_61": (221, ""),
    "SDLK_WORLD_62": (222, ""),
    "SDLK_WORLD_63": (223, ""),
    "SDLK_WORLD_64": (224, ""),
    "SDLK_WORLD_65": (225, ""),
    "SDLK_WORLD_66": (226, ""),
    "SDLK_WORLD_67": (227, ""),
    "SDLK_WORLD_68": (228, ""),
    "SDLK_WORLD_69": (229, ""),
    "SDLK_WORLD_7": (167, ""),
    "SDLK_WORLD_70": (230, ""),
    "SDLK_WORLD_71": (231, ""),
    "SDLK_WORLD_72": (232, ""),
    "SDLK_WORLD_73": (233, ""),
    "SDLK_WORLD_74": (234, ""),
    "SDLK_WORLD_75": (235, ""),
    "SDLK_WORLD_76": (236, ""),
    "SDLK_WORLD_77": (237, ""),
    "SDLK_WORLD_78": (238, ""),
    "SDLK_WORLD_79": (239, ""),
    "SDLK_WORLD_8": (168, ""),
    "SDLK_WORLD_80": (240, ""),
    "SDLK_WORLD_81": (241, ""),
    "SDLK_WORLD_82": (242, ""),
    "SDLK_WORLD_83": (243, ""),
    "SDLK_WORLD_84": (244, ""),
    "SDLK_WORLD_85": (245, ""),
    "SDLK_WORLD_86": (246, ""),
    "SDLK_WORLD_87": (247, ""),
    "SDLK_WORLD_88": (248, ""),
    "SDLK_WORLD_89": (249, ""),
    "SDLK_WORLD_9": (169, ""),
    "SDLK_WORLD_90": (250, ""),
    "SDLK_WORLD_91": (251, ""),
    "SDLK_WORLD_92": (252, ""),
    "SDLK_WORLD_93": (253, ""),
    "SDLK_WORLD_94": (254, ""),
    "SDLK_WORLD_95": (255, ""),
    "SDLK_X": (120, "DIK_X", 45, sdl2.SDL_SCANCODE_X),
    "SDLK_Y": (121, "DIK_Y", 21, sdl2.SDL_SCANCODE_Y),
    "SDLK_Z": (122, "DIK_Z", 44, sdl2.SDL_SCANCODE_Z),
}

sdl_key_code_to_name = {}
for key, value in key_table.items():
    sdl_key_code_to_name[value[0]] = key



import os
import subprocess
import threading
import traceback
import weakref
from configparser import ConfigParser
from io import TextIOWrapper

import fsui
from fsbc.application import Application
from fsbc.system import platform
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.fsuaedevicehelper import FSUAEDeviceHelper
from fsgs.input.inputdevice import InputDevice
from launcher.devicemanager import DeviceManager
from launcher.i18n import gettext
from launcher.ui.skin import Skin
from launcher.ui.widgets import CloseButton
from workspace.ui.theme import WorkspaceTheme


class JoystickConfigWindow(fsui.Window):
    def __init__(self, parent, device_name):
        title = gettext("Configure {device_name}").format(
            device_name=device_name
        )
        super().__init__(
            parent,
            title=title,
            minimizable=False,
            maximizable=False,
            separator=False,
        )
        self.theme = WorkspaceTheme.instance()
        self.layout = fsui.VerticalLayout()

        self.image = fsui.Image("workspace:res/gamepad-config.png")
        self.joystick_panel = fsui.ImageView(self, self.image)
        self.layout.add(self.joystick_panel)

        if Skin.fws():
            from workspace.ui import TitleSeparator

            separator = TitleSeparator(self)
            self.layout.add(separator, fill=True)

        panel = fsui.Panel(self)
        self.layout.add(panel, fill=True)

        panel.layout = fsui.HorizontalLayout()
        panel.layout.padding = 20

        self.device_type_ids = [
            "",
            "gamepad",
            "joystick",
            # "flightstick",
            "other",
        ]
        self.device_type_labels = [
            gettext("Choose Type"),
            gettext("Gamepad"),
            gettext("Digital Joystick"),
            # gettext("Flight Stick"),
            gettext("Other Device"),
        ]

        self.type_field = fsui.Choice(panel, self.device_type_labels)
        self.type_field.changed.connect(self.on_change)
        panel.layout.add(self.type_field)

        panel.layout.add(
            fsui.PlainLabel(panel, gettext("Make:")), margin_left=20
        )
        self.make_field = fsui.TextField(panel)
        self.make_field.set_min_width(140)
        self.make_field.changed.connect(self.on_change)
        panel.layout.add(self.make_field, margin_left=10)

        panel.layout.add(
            fsui.PlainLabel(panel, gettext("Model:")), margin_left=20
        )
        self.model_field = fsui.TextField(panel)
        self.model_field.changed.connect(self.on_change)
        panel.layout.add(self.model_field, expand=True, margin_left=10)

        self.save_button = fsui.Button(panel, gettext("Save"))
        self.save_button.activated.connect(self.on_save_button)
        panel.layout.add(self.save_button, margin_left=20)

        if self.window.theme.has_close_buttons:
            self.close_button = CloseButton(panel)
            panel.layout.add(self.close_button, margin_left=10)

        self.device_name = device_name
        existing_config = self.read_existing_config()

        self.button_panels = []
        for x, y, direction, name in BUTTONS:
            b = MappingButton(self.joystick_panel, (x, y + 4), direction, name)
            self.button_panels.append(b)
            if name in existing_config:
                b.event_name = existing_config[name]

        self.save_button.disable()
        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

        self.stopped = False
        self.current_state = {}
        self.initial_state = {}
        self.map_key_name = None

        fsui.call_later(100, self.on_timer_callback)
        thread = threading.Thread(
            target=event_thread,
            name="JoystickConfigEventThread",
            args=(self.device_name, weakref.ref(self)),
        )
        thread.start()

    def __del__(self):
        print("JoystickConfigWindow.__del__")

    def on_change(self):
        self.save_button.enable()

    def on_close(self):
        print("on_close")
        self.stopped = True

    def on_save_button(self):
        self.save_config()
        self.save_button.disable()

    def set_information(self, device_type, device_make, device_model):
        print(
            "set_information",
            repr(device_type),
            repr(device_make),
            repr(device_model),
        )
        for i, d_type in enumerate(self.device_type_ids):
            print(d_type, device_type)
            if d_type == device_type:
                self.type_field.set_index(i)
                break
        else:
            self.type_field.set_index(0)
        self.make_field.set_text(device_make)
        self.model_field.set_text(device_model)

    def read_existing_config(self):
        config_file = self.get_load_path()
        if config_file is not None:
            with open(config_file, "r", encoding="UTF-8") as f:
                return self.read_existing_config_from_stream(f)
        try:
            stream = InputDevice.get_builtin_config_for_device_guid(
                self.get_joystick_guid()
            )
            print("found builtin config for", self.get_joystick_guid())
        except LookupError:
            return {}
        else:
            with TextIOWrapper(stream, "UTF-8") as f:
                return self.read_existing_config_from_stream(f)

    def read_existing_config_from_stream(self, f):
        existing_config = {}
        parser = ConfigParser()
        try:
            parser.read_file(f)
            if parser.has_section("default"):
                for key in parser.options("default"):
                    value = parser.get("default", key)
                    existing_config[value] = key
            if parser.has_section("device"):
                device_type = parser.get("device", "type", fallback="")
                device_make = parser.get("device", "make", fallback="")
                device_model = parser.get("device", "model", fallback="")
                self.set_information(device_type, device_make, device_model)
        except Exception:
            # ignore failures loading the config
            traceback.print_exc()
        return existing_config

    # def __closed(self):
    #     print("JoystickConfigWindow.__closed")
    #     self.stopped = True
    #     # self.destroy()
    #     # self.button_panels = []

    def map_event(self, name):
        self.map_key_name = name
        for panel in self.button_panels:
            if self.map_key_name == panel.key_name:
                panel.text = "use joystick"
                panel.refresh()
            elif panel.text:
                panel.text = ""
                panel.refresh()
        self.initial_state = self.get_state()

    def get_state(self):
        return self.current_state.copy()

    def set_result(self, event_name):
        for panel in self.button_panels:
            if self.map_key_name == panel.key_name:
                panel.event_name = event_name
            elif panel.event_name == event_name:
                # remove event from other panel(s)
                panel.event_name = None
            panel.text = ""
            panel.refresh()

        self.map_key_name = None
        self.on_change()

    def on_timer_callback(self):
        if self.map_key_name:
            state = self.get_state()
            for key, value in self.initial_state.items():
                # we check initial state because some axes may have non-zero
                # rest values... (i.e. full negative)
                if state[key] and state[key] != self.initial_state[key]:
                    self.set_result(key)
        if self.stopped:
            return
        # continue timer
        fsui.call_later(100, self.on_timer_callback)

    def get_joystick_id(self):
        device_name = self.device_name.rsplit("#", 1)[0]
        buttons = DeviceManager.joystick_buttons(self.device_name)
        axes = DeviceManager.joystick_axes(self.device_name)
        hats = DeviceManager.joystick_hats(self.device_name)
        balls = DeviceManager.joystick_balls(self.device_name)
        name_lower = device_name.lower()
        name = ""
        for c in name_lower:
            if c in "abcdefghijklmnopqrstuvwxyz0123456789":
                name = name + c
            else:
                if not name.endswith("_"):
                    name += "_"
        name = name.strip("_")
        return "{0}_{1}_{2}_{3}_{4}_{5}".format(
            name, buttons, axes, hats, balls, platform
        )

    def get_joystick_guid(self):
        return DeviceManager.joystick_guid(self.device_name)

    def get_load_path(self):
        path = self.get_save_path(
            self.get_joystick_guid() + ".fs-uae-controller"
        )
        if os.path.exists(path):
            print("JoystickConfigWindow.get_load_path found", path)
            return path
        path = self.get_save_path(self.get_joystick_id() + ".conf")
        if os.path.exists(path):
            print("JoystickConfigWindow.get_load_path found", path)
            return path
        return None

    @staticmethod
    def get_save_path(file_name):
        dest = FSGSDirectories.get_controllers_dir()
        if not os.path.exists(dest):
            os.makedirs(dest)
        dest = os.path.join(dest, file_name)
        return dest

    def save_config(self):
        device_name = self.device_name.rsplit("#", 1)[0]
        device_type = self.device_type_ids[self.type_field.get_index()]
        device_make = self.make_field.get_text().strip()
        device_model = self.model_field.get_text().strip()
        guid = DeviceManager.joystick_guid(self.device_name)
        buttons = DeviceManager.joystick_buttons(self.device_name)
        axes = DeviceManager.joystick_axes(self.device_name)
        hats = DeviceManager.joystick_hats(self.device_name)
        balls = DeviceManager.joystick_balls(self.device_name)
        config = [
            "[fs-uae-controller]",
            "name = {}".format(device_name),
            "platform = {}".format(platform),
            "",
            "[device]",
            "make = {}".format(device_make),
            "model = {}".format(device_model),
            "type = {}".format(device_type),
            "",
            "[sdl]",
            "guid = {}".format(guid),
            "buttons = {}".format(buttons),
            "hats = {}".format(hats),
            "axes = {}".format(axes),
            "balls = {}".format(balls),
            "",
            "[default]",
            "include = universal_gamepad",
        ]
        button_config = []
        for i, button in enumerate(BUTTONS):
            panel = self.button_panels[i]
            if panel.event_name:
                button_config.append(
                    "{0} = {1}".format(panel.event_name, button[3])
                )
        config.extend(sorted(button_config))
        with open(
            self.get_save_path(self.get_joystick_id() + ".conf"),
            "w",
            encoding="UTF-8",
        ) as f:
            for line in config:
                f.write(line)
                f.write("\n")
        if len(guid) == 32:
            with open(
                self.get_save_path(guid + ".fs-uae-controller"),
                "w",
                encoding="UTF-8",
            ) as f:
                for line in config:
                    f.write(line)
                    f.write("\n")


class MappingButton(fsui.Panel):
    def __init__(self, parent, position, direction, name):
        super().__init__(parent)

        size = (120, 22)
        self.set_size(size)
        if direction < 0:
            position = (position[0] - size[0], position[1])
        self.set_position(position)

        self.key_name = name
        self.event_name = None
        self.text = ""
        self.direction = direction

        self.set_hand_cursor()
        self.set_background_color(fsui.Color(0xFF, 0xFF, 0xFF))

    def on_left_down(self):
        print("on_left_down")
        self.get_window().map_event(self.key_name)

    def on_paint(self):
        dc = self.create_dc()
        dc.set_font(self.get_font())  # SetFont(self.GetFont())
        if self.text:
            text = self.text
            dc.set_text_color(fsui.Color(0x00, 0x80, 0x00))
        elif self.event_name:
            text = self.event_name
            dc.set_text_color(fsui.Color(0x80, 0x80, 0x80))
        else:
            text = "click to configure"
            dc.set_text_color(fsui.Color(0xFF, 0x00, 0x00))
        tw, th = dc.measure_text(text)
        y = (self.get_size()[1] - th) // 2
        if self.direction > 0:
            x = 4
        else:
            x = self.get_size()[0] - 4 - tw
        dc.draw_text(text, x, y)


def event_thread(device_name, window_ref):
    process = FSUAEDeviceHelper.start_with_args(
        [device_name], stdout=subprocess.PIPE
    )

    # while not self.stopped and \
    while not Application.instance().stopping():
        # print("stopping?", Application.instance().stopping())
        window = window_ref()
        if window is None:
            print("window was None")
            break

        line = process.stdout.readline()
        line = line.decode("UTF-8", errors="replace")
        line = line.strip()
        parts = line.split(" ")
        if len(parts) < 2:
            continue
        type_ = parts[1]
        states = parts[2:]
        update = window.current_state

        if type_ == "buttons":
            for i, state in enumerate(states):
                state = int(state)
                update["button_{0}".format(i)] = state

        elif type_ == "axes":
            for i, state in enumerate(states):
                state = int(state)
                update["axis_" + str(i) + "_pos"] = state > 20000
                update["axis_" + str(i) + "_neg"] = state < -20000

        elif type_ == "hats":
            for i, state in enumerate(states):
                state = int(state)
                update["hat_" + str(i) + "_left"] = state & HAT_LEFT
                update["hat_" + str(i) + "_right"] = state & HAT_RIGHT
                update["hat_" + str(i) + "_up"] = state & HAT_UP
                update["hat_" + str(i) + "_down"] = state & HAT_DOWN

        # make sure references are freed in every loop iteration, so the
        # Window will get refcount 0 when it is closed.
        del update
        del window

    process.kill()


HAT_UP = 0x01
HAT_RIGHT = 0x02
HAT_DOWN = 0x04
HAT_LEFT = 0x08
BUTTONS = [
    (160, 240, -1, "dpad_left"),
    (160, 160, -1, "dpad_right"),
    (160, 200, -1, "dpad_up"),
    (160, 280, -1, "dpad_down"),
    (160, 400, -1, "lstick_left"),
    (320, 400, -1, "lstick_right"),
    (160, 360, -1, "lstick_up"),
    (160, 440, -1, "lstick_down"),
    (320, 440, -1, "lstick_button"),
    (480, 400, 1, "rstick_left"),
    (640, 400, 1, "rstick_right"),
    (640, 360, 1, "rstick_up"),
    (640, 440, 1, "rstick_down"),
    (480, 440, 1, "rstick_button"),
    (640, 160, 1, "west_button"),
    (640, 200, 1, "north_button"),
    (640, 240, 1, "east_button"),
    (640, 280, 1, "south_button"),
    (320, 80, -1, "select_button"),
    (480, 80, 1, "start_button"),
    (480, 40, 1, "menu_button"),
    (160, 40, -1, "left_shoulder"),
    (160, 80, -1, "left_trigger"),
    (640, 40, 1, "right_shoulder"),
    (640, 80, 1, "right_trigger"),
]
HELP_TEXT = """
INSTRUCTIONS

The joysticks listed are those connected when you started the program.
If you connect more, you must restart the program!

Your gamepad may not look exactly like this, so just try to map the buttons
as closely as possibly.

Some gamepads do not have a "menu" button or similar, in which case you can
skip configuring this.

Some gamepads have the d-pad and left stick physically swapped. This is not
a problem, just map the d-pad buttons against the d-pad etc.

Left and right trigger buttons are located *below* left and right shoulder
buttons.
"""


# def application(uri, args):
#     fake_uri = uri + repr(args)
#     if not raise_window(fake_uri):
#         window = JoystickConfigWindow(args[0])
#         window.show()
#         register_window(fake_uri, window)



import hashlib
import os
from binascii import unhexlify

from fsgs import Option
from fsgs.drivers.messdriver import MessDriver
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

A7800_PLATFORM_ID = "a7800"
A7800_PLATFORM_NAME = "Atari 7800"
A7800_MODEL_NTSC = "ntsc"
A7800_MODEL_PAL = "pal"
A7800_JOYSTICK = {
    "type": "joystick",
    "description": "Joystick",
    "mapping_name": "atari7800",
}
NO_CONTROLLER_TYPE = "none"
NO_CONTROLLER = {
    "type": NO_CONTROLLER_TYPE,
    "description": "None",
    "mapping_name": "",
}
A7800_PORTS = [
    {
        "description": "Port 1",
        "types": [A7800_JOYSTICK, NO_CONTROLLER],
        "type_option": "a7800_port_1_type",
        "device_option": "a7800_port_1",
    },
    {
        "description": "Port 2",
        "types": [A7800_JOYSTICK, NO_CONTROLLER],
        "type_option": "a7800_port_2_type",
        "device_option": "a7800_port_2",
    },
]
# noinspection SpellCheckingInspection
A7800_ROMS = {
    "d9d134bb6b36907c615a594cc7688f7bfcef5b43": "7800.u7",
    "14584b1eafe9721804782d4b1ac3a4a7313e455f": "c300558-001a.u7",
}

# noinspection SpellCheckingInspection
A7800P_ROMS = {"5a140136a16d1d83e4ff32a19409ca376a8df874": "7800pal.rom"}


class Atari7800PlatformHandler(Platform):
    PLATFORM_NAME = A7800_PLATFORM_NAME

    def driver(self, fsgc):
        return Atari7800MameDriver(fsgc)

    def loader(self, fsgc):
        return Atari7800Loader(fsgc)


class Atari7800Loader(SimpleLoader):
    def load_extra(self, values):
        self.config[Option.A7800_MODEL] = values["a7800_model"]
        self.config[Option.A7800_A78_HEADER] = values["a78_header"]
        self.config[Option.A7800_PORT_1_TYPE] = values["a7800_port_1_type"]
        self.config[Option.A7800_PORT_2_TYPE] = values["a7800_port_2_type"]


class Atari7800MameDriver(MessDriver):
    PORTS = A7800_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = Atari7800Helper(self.options)

    def prepare(self):
        super().prepare()
        # self.emulator.args.extend(
        #     ["-lightgun", "-lightgun_device", "mouse"])

    def get_game_file(self, config_key="cartridge_slot"):
        return self.helper.prepare_rom(self)

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, port):
        return {
            "START": "P#_START",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
        }

    def mess_romset(self):
        if self.helper.model() == A7800_MODEL_PAL:
            return "a7800p", A7800P_ROMS
        else:
            return "a7800", A7800_ROMS


class Atari7800Helper:
    def __init__(self, options):
        self.options = options

    def model(self):
        model = self.options[Option.A7800_MODEL]
        if model in ["", A7800_MODEL_NTSC]:
            return A7800_MODEL_NTSC
        elif model == A7800_MODEL_PAL:
            return A7800_MODEL_PAL
        else:
            print("[A7800] Warning: Invalid model:", model)
            return A7800_MODEL_NTSC

    def pal(self):
        return self.model() == A7800_MODEL_PAL

    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    def prepare_rom_with_stream(self, driver, input_stream, ext):
        # This should not be necessary for files found via the file database
        # and the online game database, but could be necessary for manually
        # loaded files.
        data = input_stream.read(128)
        if len(data) == 128 and data[1:10] == b"ATARI7800":
            print("[A7800] Stripping A78 header")
            data = None
        else:
            # No A78 header, include data
            pass
        sha1_obj = hashlib.sha1()
        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            header = self.options[Option.A7800_A78_HEADER]
            if header:
                assert len(header) == 128 * 2
                f.write(unhexlify(header))
                sha1_obj.update(unhexlify(header))
            if data is not None:
                f.write(data)
                sha1_obj.update(data)
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path



#!/usr/bin/env python2.7

from __future__ import absolute_import
from __future__ import print_function

import sys
import os.path
import amitools.util.KeyValue as KeyValue
import amitools.util.ByteSize as ByteSize
from amitools.fs.blkdev.DiskGeometry import DiskGeometry
from amitools.fs.blkdev.BlkDevFactory import BlkDevFactory

def main():
  a = sys.argv
  n = len(a)
  if n < 3:
    print("Usage: (detect <size|file> [options] | setup <options> | open <file> [options] | create <file> <options>)")
    print("""Options:
             size=<size>
             chs=<n>,<n>,<n>
             c=<n> h=<n> s=<n>
             algo=1|2
             """)
    return 1
  else:
    cmd = a[1]
    # detect disk geometry from given image file
    if cmd == 'detect':
      if os.path.exists(a[2]):
        # its a file
        size = os.path.getsize(a[2])
      else:
        # parse size string
        size = ByteSize.parse_byte_size_str(a[2])
      if size == None:
        print("Invalid size!")
      else:
        d = DiskGeometry()
        opts = None
        if n > 3:
          opts = KeyValue.parse_key_value_strings(a[3:])
        print("size:  ",size)
        print("opts:  ",opts)
        size = d.detect(size, opts)
        if size != None:
          print("geo:   ", d)
        else:
          print("FAILED")
    # setup a new disk geometry from options
    elif cmd == 'setup' :
      d = DiskGeometry()
      opts = KeyValue.parse_key_value_strings(a[2:])
      print("opts:  ", opts)
      size = d.setup(opts)
      if size != None:
        print("setup: ", size, ByteSize.to_byte_size_str(size))
        print("geo:   ", d)
      else:
        print("FAILED")
    # open a blkdev and detect geometry
    elif cmd == 'open':
      opts = None
      if n > 3:
        opts = KeyValue.parse_key_value_strings(a[3:])
      print("opts:   ", opts)
      f = BlkDevFactory()
      blkdev = f.open(a[2], options=opts)
      if blkdev != None:
        print("blkdev: ", blkdev.__class__.__name__)
        print("geo:    ", blkdev.get_geometry())
        blkdev.close()
      else:
        print("FAILED")
    # create a new blkdev with setup geometry
    elif cmd == 'create':
      opts = KeyValue.parse_key_value_strings(a[3:])
      print("opts:   ", opts)
      f = BlkDevFactory()
      blkdev = f.create(a[2], options=opts)
      if blkdev != None:
        print("blkdev: ",blkdev.__class__.__name__)
        print("geo:    ",blkdev.get_geometry())
        blkdev.close()
      else:
        print("FAILED")
    return 0

if __name__ == '__main__':
  sys.exit(main())



from fsgs.platform import PlatformHandler
from fsgs.platforms.dos.dosboxdosdriver import DosBoxDosDriver
from fsgs.platforms.loader import SimpleLoader


class DOSPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "DOS"

    def __init__(self):
        super().__init__()

    def get_loader(self, fsgs):
        return DOSLoader(fsgs)

    def get_runner(self, fsgs):
        return DosBoxDosDriver(fsgs)


class DOSLoader(SimpleLoader):
    def load_files(self, values):
        self.config["file_list"] = values["file_list"]
        self.config["cue_sheets"] = values["cue_sheets"]

    def load_extra(self, values):
        for key in values.keys():
            if key.startswith("dosbox_"):
                self.config[key] = values[key]
        # for key in ["dosbox_cpu_cycles", "dosbox_sblaster_sbtype",
        #             "dosbox_sblaster_sbbase", "dosbox_sblaster_irq",
        #             "dosbox_sblaster_oplrate"]:
        #     self.config[key] = values[key]

        self.config["command"] = values["command"]
        if not self.config["command"]:
            # Deprecated option
            self.config["command"] = values["hd_startup"]



#!/usr/bin/env python
from __future__ import print_function

import os
import shutil
import subprocess
import tempfile

dirs = ["arcade", "fsui", "fsbc", "fsgs", "launcher", "workspace"]
files = []
extensions = [".py"]

temp = tempfile.NamedTemporaryFile(delete=False)
for dir in dirs:
    for dir_path, dir_names, file_names in os.walk(dir):
        for file in file_names:
            name, ext = os.path.splitext(file)
            if ext not in extensions:
                continue
            path = os.path.join(dir_path, file)
            files.append(path)

with temp:
    for path in sorted(files):
        temp.write(path)
        temp.write("\n")

args = [
    "xgettext",
    "--output-dir=po",
    "--files-from=" + temp.name,
    "--keyword=_",
    "--keyword=N_",
    "--package-name=fs-uae-launcher",
    "--msgid-bugs-address=frode@fs-uae.net",
    "--copyright-holder=Frode Solheim",
    "--add-comments=#",
    "--no-location",
    "--sort-output",
    "--from-code=ISO-8859-1",
]

print(args)
p = subprocess.Popen(args)
p.wait()
shutil.move("po/messages.po", "po/messages.pot")

# for file in os.listdir("po"):
#    if file.endswith(".po"):
#        path = os.path.join("po", file)
#        print(path)
#        p = subprocess.Popen(["msgmerge",
#                              "--no-fuzzy-matching",
#                              "--sort-output",
#                              "--update",
#                              path,
#                              "po/messages.pot"])
#        p.wait()

os.system("find share -name fs-uae.mo -delete")
# os.system("make -C .. po-dist")
os.system("make mo")



import fsui
from launcher.i18n import gettext
from launcher.ui.behaviors.configbehavior import ConfigBehavior


class OverrideWarning(fsui.ImageView):
    def __init__(self, parent, name):
        fsui.ImageView.__init__(
            self, parent, fsui.Image("launcher:res/16x16/warning_2.png")
        )
        setattr(self, "on_{0}_config".format(name), self.on_config)
        ConfigBehavior(self, [name])
        text = gettext(
            "Option {name} is overridden by current configuration".format(
                name=name
            )
        )
        self.set_tool_tip(text)

    def on_config(self, value):
        self.show_or_hide(bool(value))



# this is a quick hack tp fix problems with some modules not being imported
# by py2exe / py2app otherwise

# import workspace.apps.adf_creator_app
# import workspace.apps.hdf_creator_app
# import workspace.apps.joystick_config_app
# import workspace.apps.locker_uploader
# import workspace.apps.login
# import workspace.apps.logout
# import workspace.apps.refresh

# import workspace.apps.clock_app
# import workspace.prefs.audio
# import workspace.prefs.experimental_features
# import workspace.prefs.filter
# import workspace.prefs.game_database
# import workspace.prefs.joystick
# import workspace.prefs.keyboard
# import workspace.prefs.language
# import workspace.prefs.mouse
# import workspace.prefs.netplay
# # import workspace.prefs.opengl
# import workspace.prefs.scan
# import workspace.prefs.video



class Constants:
    TAB_WIDTH = 64
    TAB_HEIGHT = 64

    # COVER_SIZE = (108, 136)
    # SCREEN_SIZE = (224, 140)
    # SCREEN_SIZE = (216, 136)
    SCREEN_SIZE = (210, 134)
    COVER_SIZE = (106, 134)

    # 2 for top border, 1 + 1 for border over and under screens
    # BOTTOM_HEIGHT = SCREEN_SIZE[1] + 20 + + 2 + 1 + 1 + 10



from arcade.glui.animation import AnimateValueBezier
from arcade.glui.navigatable import Navigatable
from arcade.glui.state import State
from arcade.glui.topmenu import TopMenu
from fsbc.application import app


class Menu(Navigatable):
    def __init__(self, title=""):
        self.navigatable = None
        self.top = TopMenu()
        # self.top_mode = False
        self.title = title
        self.parent_menu = None
        self.parents = []
        self.items = []
        self._selected_index = 0
        self.position = 0.0
        self._position_animation = None
        self.last_scroll_pos = 0
        self.last_scroll_time = 0
        self.cur_scroll_speed = 0
        self.configuration_index = 0
        self.top_menu_transition = 1.0
        self.search_text = ""

    def use_game_center_item(self):
        return app.settings["game-center:top-logo"] != "0"

    def update_path(self, menu_path):
        # FIXME: UPDATE PATH
        # return
        print("update_path", menu_path)
        if self.use_game_center_item():
            from arcade.glui.topmenu import GameCenterItem

            self.top.append_left(GameCenterItem())
        # if len(menu_path) > 0:
        from arcade.glui.topmenu import HomeItem

        self.top.append_left(HomeItem())
        last_index = len(menu_path) - 1
        for i, item in enumerate(menu_path):
            if item.path_title or last_index == i:
                print("adding item", item)
                self.top.append_left(item)
                # if last_index == i:
                #     item.enabled = False
                # else:
                #     item.enabled = True

                # FIXME: FOR NOW, ALWAYS DISABLE PATH ELEMENTS
                item.enabled = False
        self.set_default_top_selected()

    def set_default_top_selected(self):
        # for i in range(len(self.top.left) - 1, -1, -1):
        #     if self.top.left[i].enabled:
        #         self.top.set_selected_index(i)
        #         break
        i = len(self.top.left) + len(self.top.right) - 1
        self.top.set_selected_index(i)

    def add_add_item(self):
        from arcade.glui.topmenu import AddItem

        self.top.append_left(AddItem())
        self.set_default_top_selected()
        # print("FIXME: add_add_item temporarily disabled")

    def append(self, item):
        self.items.append(item)

    def remove(self, item):
        self.items.remove(item)

    def reset_position(self):
        self.position = self._selected_index
        self.configuration_index = 0

    @property
    def selected_item(self):
        # if self.top_mode:
        #     return self.top.selected_item
        return self.items[self._selected_index % len(self.items)]

    def get_selected_index(self):
        return self._selected_index

    def get_current_position(self):
        t = State.get().time
        if self.last_scroll_time == t:
            pass
        elif self.last_scroll_time - t < 1.0:
            self.cur_scroll_speed = (self.position - self.last_scroll_pos) / (
                t - self.last_scroll_time
            )
        else:
            self.cur_scroll_speed = 0
        self.last_scroll_pos = self.position
        self.last_scroll_time = t
        # print "speed", self.cur_scroll_speed

        return self.position

    def set_selected_index(self, index, immediate=False):
        self._selected_index = index
        if immediate:
            self.reset_position()
        else:
            p1 = self.position + self.cur_scroll_speed * 0.040
            self._position_animation = AnimateValueBezier(
                (self, "position"),
                self.position,
                State.get().time,
                p1,
                State.get().time + 0.040,
                index,
                State.get().time + 0.080,
                index,
                State.get().time + 0.200,
            )
        self.configuration_index = 0

    def go_left(self, count=1):
        # FIXME:
        from arcade.glui.window import set_items_brightness

        set_items_brightness(1.0, duration=0.300)
        self.set_selected_index(self._selected_index - count)

    def go_right(self, count=1):
        from arcade.glui.window import set_items_brightness

        set_items_brightness(1.0, duration=0.300)
        self.set_selected_index(self._selected_index + count)

    def activate(self):
        result = self.selected_item.activate(self)
        if result is None:
            return
        # FIXME
        from arcade.glui.window import enter_menu

        enter_menu(result)

    def render(self):
        pass

    def update(self):
        pass

    def render_transparent(self, data):
        pass

    def __getitem__(self, index):
        return self.items[index]

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return "<{2} {0} ({1} items)>".format(
            self.title, len(self), self.__class__.__name__
        )



import hashlib
import os

from fsgs import Option
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

# FIXME: Use KnownFile
# [BIOS] Game Boy Advance (World).gba
GBA_WORLD_BIOS_SHA1 = "300c20df6731a33952ded8c436f7f186d25d3492"

GBA_CONTROLLER = {
    "type": "gamepad",
    "description": "Built-in Controller",
    "mapping_name": "gameboyadvance",
}
GBA_PORTS = [
    {
        "description": "Built-in",
        "types": [GBA_CONTROLLER],
        "type_option": "gba_port_1_type",
        "device_option": "gba_port_1",
    }
]


class GameBoyAdvancePlatform(Platform):
    PLATFORM_NAME = "Game Boy Advance"

    def driver(self, fsgc):
        return GameBoyAdvanceMednafenDriver(fsgc)

    def loader(self, fsgc):
        return GameBoyAdvanceLoader(fsgc)


class GameBoyAdvanceLoader(SimpleLoader):
    def load_extra(self, values):
        self.config[Option.SRAM_TYPE] = values["sram_type"]


class GameBoyAdvanceMednafenDriver(MednafenDriver):
    PORTS = GBA_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = GameBoyAdvanceHelper(self.options)
        self.sram_type_file = None

    def prepare(self):
        print("[GBA] Mednafen GBA driver preparing...")
        super().prepare()
        self.set_mednafen_aspect(3, 2)
        # We do aspect calculation separately. Must not be done twice.
        # self.emulator.args.extend(["-snes.correct_aspect", "0"])
        # FIXME: Input ports configuration
        # FIXME: SNES model
        rom_path = self.helper.prepare_rom(self)
        self.emulator.args.append(rom_path)

        self.configure_gba_sram(rom_path)
        self.configure_gba_colormap()
        self.configure_gba_bios()

        print("[FIXME: temporarily removed support for custom sav file")
        # sav_file = os.path.splitext(rom_file)[0] + u".sav"
        # if os.path.exists(sav_file):
        #     m = hashlib.md5()
        #     with open(rom_file, "rb") as f:
        #         while True:
        #             data = f.read(16384)
        #             if not data:
        #                 break
        #             m.update(data)
        #         md5sum = str(m.hexdigest())
        #     save_name = os.path.splitext(
        #         os.path.basename(rom_file))[0] + u"." + md5sum + u".sav"
        #     dest_path = os.path.join(self.get_state_dir(), save_name)
        #     if not os.path.exists(dest_path):
        #         shutil.copy(sav_file, dest_path)

    # def init_input(self):
    #     # self.inputs = [
    #     #     self.create_input(name="Controller",
    #     #             type="gameboyadvance",
    #     #             description="Built-in Gamepad"),
    #     # ]
    #     self.set_mednafen_input_order()

    def finish(self):
        if self.sram_type_file:
            try:
                os.remove(self.sram_type_file)
            except Exception:
                # Not expected to happen, but is not critical if it does.
                print("[GBA] Could not remove SRAM type file")
        super().finish()

    def get_game_refresh_rate(self):
        # all GBA games should use a refresh rate of approx. 60.0 Hz
        # (or 30.0 Hz)
        return 59.73

    def mednafen_input_mapping(self, _):
        return {
            "A": "gba.input.builtin.gamepad.a",
            "B": "gba.input.builtin.gamepad.b",
            "L": "gba.input.builtin.gamepad.shoulder_l",
            "R": "gba.input.builtin.gamepad.shoulder_r",
            "UP": "gba.input.builtin.gamepad.up",
            "DOWN": "gba.input.builtin.gamepad.down",
            "LEFT": "gba.input.builtin.gamepad.left",
            "RIGHT": "gba.input.builtin.gamepad.right",
            "SELECT": "gba.input.builtin.gamepad.select",
            "START": "gba.input.builtin.gamepad.start",
        }

    def mednafen_rom_extensions(self):
        return [".gba"]

    def mednafen_scanlines_setting(self):
        return 33

    def mednafen_special_filter(self):
        return "nn2x"

    def mednafen_system_prefix(self):
        return "gba"

    def game_video_par(self):
        return 1.0

    def game_video_size(self):
        return 240, 160

    def configure_gba_sram(self, rom_path):
        if self.options[Option.SRAM_TYPE]:
            save_dir = self.save_handler.save_dir()
            rom_name = os.path.splitext(os.path.basename(rom_path))[0]
            type_file = os.path.join(save_dir, rom_name + ".type")
            with open(type_file, "wb") as f:
                for line in self.options[Option.SRAM_TYPE].split(";"):
                    f.write((line.strip() + "\n").encode("UTF-8"))
            self.sram_type_file = type_file

    def configure_gba_colormap(self):
        gamma = 1.3
        if self.options[Option.GBA_GAMMA]:
            try:
                gamma = float(self.options[Option.GBA_GAMMA])
            except ValueError:
                print("[GBA] WARNING: Invalid gamma value")
        self.create_colormap(os.path.join(self.home.path, "gba.pal"), gamma)
        # if os.path.exists(os.path.join(self.home.path, "gba.pal")):
        #     os.remove(os.path.join(self.home.path, "gba.pal"))
        # self.colormap_temp = self.temp_file("color.map")
        # self.create_colormap(self.colormap_temp.path, gamma)
        # self.emulator.args.extend(
        #     ["-gba.colormap", self.colormap_temp.path])

    def configure_gba_bios(self):
        uri = self.fsgc.file.find_by_sha1(GBA_WORLD_BIOS_SHA1)
        stream = self.fsgc.file.open(uri)
        if stream is not None:
            bios_temp = self.temp_file("gba_bios.bin")
            with open(bios_temp.path, "wb") as f:
                f.write(stream.read())
            self.emulator.args.extend(["-gba.bios", bios_temp.path])
        else:
            print("[GBA] WARNING: GBA BIOS not found, using HLE")

    @staticmethod
    def create_colormap(path, gamma):
        with open(path, "wb") as f:
            for x in range(32768):
                r = (x & 0x1F) << 3
                g = ((x & 0x3E0) >> 5) << 3
                b = ((x & 0x7C00) >> 10) << 3
                r /= 255.0
                g /= 255.0
                b /= 255.0
                # h, l, s = colorsys.rgb_to_hls(r, g, b)
                # l = l ** gamma
                # r, g, b = colorsys.hls_to_rgb(h, l, s)
                r = r**gamma
                g = g**gamma
                b = b**gamma
                f.write(bytes([int(r * 255)]))
                f.write(bytes([int(g * 255)]))
                f.write(bytes([int(b * 255)]))

    def get_game_file(self, config_key="cartridge_slot"):
        return None


class GameBoyAdvanceHelper:
    def __init__(self, options):
        self.options = options

    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    def prepare_rom_with_stream(self, driver, input_stream, ext):
        sha1_obj = hashlib.sha1()
        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path



import datetime
import hashlib
import io
import os
import time

from fsgs.context import fsgs
from fsgs.Database import Database
from fsgs.filedatabase import FileDatabase
from fsgs.FSGSDirectories import FSGSDirectories
from launcher.configuration_scanner import ConfigurationScanner
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.ui.IconButton import IconButton


class SaveButton(IconButton):
    def __init__(self, parent):
        super().__init__(parent, "save_button.png")
        self.set_tooltip(gettext("Save Config"))
        LauncherConfig.add_listener(self)
        self.on_config("__changed", LauncherConfig.get("__changed"))

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key == "__changed":
            self.enable(value == "1")

    def on_activate(self):
        print("SaveButton.on_activate")
        # try:
        self.save_config()
        # except Exception:
        #     # FIXME: notify user
        #     pass

    @staticmethod
    def save_config():
        print("SaveButton.save_config")
        database = Database.get_instance()

        name = LauncherSettings.get("config_name").strip()
        if not name:
            print("no config_name")
            # FIXME: notify user
            return

        file_name = name + ".fs-uae"
        path = os.path.join(
            FSGSDirectories.get_configurations_dir(), file_name
        )
        with io.open(path, "w", encoding="UTF-8") as f:
            f.write("# FS-UAE configuration saved by FS-UAE Launcher\n")
            f.write(
                "# Last saved: {0}\n".format(
                    datetime.datetime.today().strftime("%Y-%m-%d %H:%M:%S")
                )
            )
            f.write("\n[fs-uae]\n")
            keys = sorted(fsgs.config.values.keys())
            for key in keys:
                value = LauncherConfig.get(key)
                if key.startswith("__"):
                    continue
                if key in LauncherConfig.no_save_keys_set:
                    continue
                # elif key == "joystick_port_2_mode" and value == "nothing":
                #     continue
                # elif key == "joystick_port_3_mode" and value == "nothing":
                #     continue
                if value == LauncherConfig.default_config.get(key, ""):
                    continue
                if value:
                    f.write("{0} = {1}\n".format(key, value))

        # scanner = ConfigurationScanner()
        # search = ConfigurationScanner.create_configuration_search(name)
        # name = scanner.create_configuration_name(name)
        # print("adding", path)
        # # deleting the path from the database first in case it already exists
        # database.delete_configuration(path=path)
        # database.delete_file(path=path)
        # database.add_file(path=path)
        # database.add_configuration(
        #     path=path, uuid="", name=name, scan=0, search=search)

        file_database = FileDatabase.get_instance()
        scanner = ConfigurationScanner()
        print("[save config] adding config", path)
        file_database.delete_file(path=path)
        with open(path, "rb") as f:
            sha1 = hashlib.sha1(f.read()).hexdigest()
        file_database.add_file(path=path, sha1=sha1)

        game_id = database.add_configuration(
            path=path, name=scanner.create_configuration_name(name)
        )
        database.update_game_search_terms(
            game_id, scanner.create_search_terms(name)
        )

        database.commit()
        file_database.commit()

        LauncherSettings.set("__config_refresh", str(time.time()))
        # Settings.set("config_changed", "0")
        LauncherConfig.set("__changed", "0")



import filecmp
import os
import shutil

from fsbc.system import windows
from fsbc.task import current_task
from fsgs.drivers.gamedriver import Emulator, GameDriver
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.input.mapper import InputMapper
from fsgs.knownfiles import KnownFilePath
from fsgs.option import Option
from fsgs.plugins.pluginmanager import PluginManager

"""

FIXME: Check screenshot saving

FIXME: It would be nice to be able to go back in the on-screen menu (UI_CANCEL)
without quitting the emu when the menu is gone. Introduce a separate quit
action? And then also map a joystick button back?
"""


class MameCheatFile(KnownFilePath):
    @property
    def path(self):
        return os.path.join(
            FSGSDirectories.get_data_dir(), "MAME", "Cheats", "cheat.7z"
        )


class MameArtworkDirectory(KnownFilePath):
    # FIXME: KnownDirectory
    @property
    def path(self):
        return os.path.join(FSGSDirectories.get_data_dir(), "MAME", "Artwork")


mame_cheat_file = MameCheatFile()
mame_artwork_dir = MameArtworkDirectory()


# noinspection PyAttributeOutsideInit
class MameDriver(GameDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator = Emulator("mame-fs")
        self.mame_init()

    def mame_emulator_name(self):
        return "mame"

    def mame_init_input(self):
        """Override in inherited classes"""
        pass

    def mame_romset(self):
        """Override in inherited classes"""
        raise NotImplementedError()

    def init_input(self):
        self.mame_init_input()
        if windows:
            self.input_device_order = "DINPUT8"

    def install_mame_hash_file(self, name):
        # FIXME: Better to find data file based on path/provides rather than
        # hardcoding plugin name, but...
        plugin = PluginManager.instance().plugin("MAME-FS")
        src = plugin.data_file_path("hash/" + name)
        hash_dir = os.path.join(self.cwd.path, "hash")
        if not os.path.exists(hash_dir):
            os.makedirs(hash_dir)
        dst = os.path.join(hash_dir, name)
        # FIXME: Check if file is the same (no need to copy)
        if os.path.exists(dst) and filecmp.cmp(src, dst):
            print("[A5200] Hash file", name, "already in place")
            return
        print("[A5200] Installing hash/" + name)
        shutil.copy2(src, dst)

    def prepare(self):
        self.configure()
        self.mame_prepare()

    def create_mame_layout(self):
        paths = self.emulator_skin_paths()
        artwork = os.path.join(self.cwd.path, "artwork", self.romset)
        if not os.path.exists(artwork):
            os.makedirs(artwork)

        if self.stretching() == self.STRETCH_FILL_SCREEN or not self.bezel():
            shutil.rmtree(os.path.join(self.cwd.path, "artwork"))
            return

        # FIXME: filecmp?
        self.prepare_emulator_skin(
            paths={
                "left": os.path.join(artwork, "left.png"),
                "right": os.path.join(artwork, "right.png"),
            }
        )
        # shutil.move(paths["left"], )
        # shutil.move(paths["right"], )

        # FIXME: With no bezel, we should still use a black bezel to
        # hide screen stretching

        # FIXME: On taller displays than 16:9, the bezel can cause the
        # graphics to shrink (fit-in). Is it possible to display the
        # bezel

        if self.options["orientation"] == "vertical":
            # noinspection SpellCheckingInspection
            layout = """<mamelayout version="2">
                <element name="left">
                    <image file="left.png" />
                </element>
                <element name="right">
                    <image file="right.png" />
                </element>
                <view name="FSGS Bezel">
                    <screen index="0">
                        <bounds x="0" y="0" width="810" height="1080" />
                    </screen>
                    <bezel element="left">
                        <bounds x="-160" y="0" width="160" height="1080" />
                    </bezel>
                    <bezel element="right">
                        <bounds x="810" y="0" width="160" height="1080" />
                    </bezel>
                </view></mamelayout>"""
        else:
            # noinspection SpellCheckingInspection
            layout = """<mamelayout version="2">
                <element name="left">
                    <image file="left.png" />
                </element>
                <element name="right">
                    <image file="right.png" />
                </element>
                <view name="FSGS Bezel">
                    <screen index="0">
                        <bounds x="0" y="0" width="1440" height="1080" />
                    </screen>
                    <bezel element="left">
                        <bounds x="-160" y="0" width="160" height="1080" />
                    </bezel>
                    <bezel element="right">
                        <bounds x="1440" y="0" width="160" height="1080" />
                    </bezel>
                </view></mamelayout>"""
        with open(os.path.join(artwork, "default.lay"), "w") as f:
            f.write(layout)

    def mame_init(self):
        # override in subclasses
        pass

    def mame_prepare(self):
        pass

    def configure(self):
        self.configure_romset()

        self.default_xml = [
            """\ufeff<?xml version="1.0"?>
<!-- This file is autogenerated; comments and unknown tags will be stripped -->
<mameconfig version="10">
    <system name="default">\n"""
        ]
        self.game_xml = [
            """\ufeff<?xml version="1.0"?>
<!-- This file is autogenerated; comments and unknown tags will be stripped -->
<mameconfig version="10">
    <system name="{0}">\n""".format(self.romset)
        ]

        self.emulator.args.extend(["-skip_gameinfo"])
        # state_dir = self.get_state_dir()
        base_save_dir = self.emulator_state_dir(self.mame_emulator_name())
        emu_save_dir = base_save_dir
        if hasattr(self, "save_handler"):
            # save_dir = self.save_handler.save_dir()
            base_save_dir = self.save_handler.save_dir()
            emu_save_dir = self.save_handler.emulator_save_dir()
            base_save_dir = base_save_dir + os.sep
        emu_save_dir = emu_save_dir + os.sep

        emu_state_dir = self.emulator_state_dir(self.mame_emulator_name())
        if hasattr(self, "save_handler"):
            emu_state_dir = self.save_handler.emulator_state_dir()
        emu_state_dir = emu_state_dir + os.sep

        self.cwd_dir = self.create_temp_dir("mame-cwd")
        self.cfg_dir = self.create_temp_dir("mame-cfg")
        self.roms_dir = self.create_temp_dir("mame-roms")
        rom_path = self.roms_dir.path
        assert self.romset
        system_rom_path = os.path.join(rom_path, self.romset)
        os.makedirs(system_rom_path)

        # for sha1, name in self.romset_files:
        for name, sha1 in self.romset_files:
            file_uri = self.fsgs.file.find_by_sha1(sha1)
            current_task.set_progress("Preparing ROM {name}".format(name=name))
            input_stream = self.fsgs.file.open(file_uri)
            if input_stream is None:
                raise Exception("Cannot not find required ROM " + repr(name))
            path = os.path.join(system_rom_path, name)
            with open(path, "wb") as f:
                f.write(input_stream.read())

        # shutil.copy(self.get_game_file(),
        #             os.path.join(rom_path, self.romset + '.zip'))

        # MAME uses ; as path separator on all systems, apparently
        try:
            # self.args.extend(["-rompath", self.bios_dir()])
            rom_path = rom_path + ";" + self.mame_get_bios_dir()
        except:
            pass
        # rom_path = rom_path + ";" + os.path.dirname(self.get_game_file())

        # rom_path = rom_path + os.pathsep + os.path.dirname(
        #        self.get_game_file())
        self.add_arg("-rompath", rom_path)

        # copy initial nvram disk, if any, to nvram dir
        # p, e = os.path.splitext(self.get_game_file())
        # initram = p + ".nv"
        # if os.path.exists(initram):
        #    shutil.copy(initram, os.path.join(state_dir, self.romset + '.nv'))

        game_cfg_file = os.path.join(
            self.cfg_dir.path, "{romset}.cfg".format(romset=self.romset)
        )
        self.emulator.args.extend(["-cfg_directory", self.cfg_dir.path])
        # self.add_arg("-memcard_directory", state_dir)
        # self.add_arg("-hiscore_directory", state_dir)
        # FIXME: What is this?
        self.emulator.args.extend(["-diff_directory", emu_state_dir])

        self.emulator.args.extend(["-nvram_directory", emu_save_dir])

        # We not not need nor want system-specific sub-directories since we
        # already have unique save directories per game variant.
        self.emulator.args.extend(["-statename", "MAME"])
        self.emulator.args.extend(["-state_directory", base_save_dir])

        self.emulator.args.extend(
            ["-snapshot_directory", self.screenshots_dir()]
        )
        self.emulator.args.extend(
            ["-snapname", "{0}-%i".format(self.screenshots_name())]
        )

        # self.change_handler = GameChangeHandler(self.cfg_dir.path)
        # self.change_handler.init(
        #     os.path.join(self.get_state_dir(), "cfg"))

        self.configure_input()
        self.configure_video()

        self.game_xml.append("    </system>\n")
        self.game_xml.append("</mameconfig>\n")
        with open(game_cfg_file, "wb") as f:
            f.write("".join(self.game_xml).encode("UTF-8"))
            print("")
            print("")
            print("".join(self.game_xml))
            print("")
            print("")

        self.default_xml.append("    </system>\n")
        self.default_xml.append("</mameconfig>\n")
        with open(os.path.join(self.cfg_dir.path, "default.cfg"), "wb") as f:
            f.write("".join(self.default_xml).encode("UTF-8"))
            print("")
            print("")
            print("".join(self.default_xml))
            print("")
            print("")

        if self.use_doubling():
            self.emulator.args.extend(["-prescale", "2"])

        if self.use_smoothing():
            self.emulator.args.append("-filter")
        else:
            self.emulator.args.append("-nofilter")

        cheats_file_path = mame_cheat_file.path
        if not os.path.exists(cheats_file_path):
            # Data/MAME/Cheat/cheat.7z not found, trying plugin instead.
            cheats_file_path = self.cheats_file("MAME/cheat.7z")
        if cheats_file_path:
            print("[MAME] Using cheats file:".format(cheats_file_path))
            self.emulator.args.extend(
                ["-cheatpath", cheats_file_path[:-3]]
            )  # Stripping .7z
            self.emulator.args.extend(["-cheat"])
        else:
            print("[MAME] No cheats file found")

        self.emulator.args.append(self.romset)
        self.mame_configure()

    def mame_configure(self):
        # override in subclasses
        pass

    def configure_romset(self):
        self.romset, self.romset_files = self.mame_romset()
        try:
            # convert from dict to list
            # noinspection PyUnresolvedReferences
            self.romset_files = list(self.romset_files.items())
        except AttributeError:
            # already a list
            pass

    def configure_input(self):
        # return
        # FIXME: ignoring input
        self.default_xml.append("        <input>\n")
        self.game_xml.append("        <input>\n")
        ports = {}
        # for i, input in enumerate(self.inputs):
        for i, port in enumerate(self.ports):
            input_mapping = self.mame_input_mapping(i)

            mapper = MameInputMapper(port, input_mapping)
            for key, value in mapper.items():
                print("---->", key, value)
                if isinstance(key, tuple):
                    key, type = key
                else:
                    type = "standard"
                if "type=" not in key:
                    key = 'type="{0}"'.format(key)
                key = key.replace("#", str(i + 1))
                # if '/' in key:
                #     key, tag = key.split('/')
                # else:
                #     tag = None
                # if ':' in key:
                #     key, type = key.split(':')
                # else:
                #     type = 'standard'
                if "AD_STICK" in key:  # and type == 'standard':
                    analog_axis = True
                else:
                    analog_axis = False
                if analog_axis and "AXIS" in value:
                    value = value[: value.index("AXIS") + 4]
                    # remove increment / decrement type, set type
                    # to standard since this is an analog axis
                    type = "standard"
                ports.setdefault(key, {}).setdefault(type, set()).add(value)
        for key, port in ports.items():
            # key, tag = key
            if "tag=" in key:
                xml = self.game_xml
                # xml.append(
                #         '            <port tag="{tag}" '
                #         'type="{key}" mask="1" default="0">'
                #         '\n'.format(tag=tag, key=key))
                # xml.append(
                #         '            <port {key}>\n'.format(key=key))
            else:
                xml = self.default_xml
                # xml.append(
                #         '            <port type="{key}">\n'.format(key=key))
                # xml.append(
                #         '            <port {key}>\n'.format(key=key))
            xml.append("            <port {key}>\n".format(key=key))
            for type, values in port.items():
                xml.append(
                    '                <newseq type="{type}">\n'.format(
                        type=type
                    )
                )
                xml.append("                    ")
                for i, value in enumerate(values):
                    if i > 0:
                        xml.append(" OR ")
                    xml.append(value)
                xml.append("\n                </newseq>\n")
            xml.append("            </port>\n")

        self.add_default_shortcuts(self.mame_shortcuts())

        self.default_xml.append("        </input>\n")
        self.game_xml.append("        </input>\n")

    def add_default_shortcuts(self, shortcuts):
        xml = self.default_xml
        for key, values in shortcuts.items():
            xml.append('            <port type="{key}">\n'.format(key=key))
            xml.append('                <newseq type="standard">\n')
            xml.append("                    ")
            for i, value in enumerate(values):
                if i > 0:
                    xml.append(" OR ")
                xml.append(value)
            xml.append("\n                </newseq>\n")
            xml.append("            </port>\n")

    def mame_shortcuts(self):
        shortcuts = {
            "TOGGLE_KEEP_ASPECT": ["KEYCODE_LALT KEYCODE_A"],
            "UI_CANCEL": ["KEYCODE_LALT KEYCODE_Q"],
            "UI_PAUSE": ["KEYCODE_LALT KEYCODE_P"],
            "UI_CONFIGURE": ["KEYCODE_F12"],
            "UI_SNAPSHOT": ["KEYCODE_LALT KEYCODE_S"],
            "UI_THROTTLE": ["KEYCODE_LALT KEYCODE_W"],
            "UI_TOGGLE_UI": ["KEYCODE_LALT KEYCODE_K", "KEYCODE_SCRLOCK"],
        }
        return shortcuts

    def configure_video(self):
        if windows and False:
            self.emulator.args.extend(["-video", "d3d"])
        else:
            self.emulator.args.extend(["-video", "opengl"])
        if self.use_fullscreen():
            self.emulator.args.extend(["-nowindow"])
        else:
            self.emulator.args.extend(["-window", "-nomaximize"])
            if self.use_stretching():
                self.emulator.args.extend(["-resolution", "960x540"])
            else:
                # FIXME: Square pixels and no stretch... how to open window
                # at appropriate size? (with and without pixel doubling?
                pass

        if self.stretching() == self.STRETCH_FILL_SCREEN:
            self.emulator.args.extend(["-nokeepaspect"])
        else:
            self.emulator.args.extend(["-keepaspect"])

        # FIXME: Square pixels and no stretch...

        self.game_xml.append("        <video>\n")
        # if self.stretching() == self.NO_STRETCHING:
        #     self.game_xml.append('            <target index="0" '
        #                          'view="Pixel Aspect (" />\n')

        if self.options[Option.MAME_ARTWORK] == "1":
            # Do not use scaling options
            self.emulator.args.extend(["-artpath", mame_artwork_dir.path])
            pass
        else:
            self.game_xml.append('            <screen index="0" ')
            ox, oy, sx, sy = self.mame_offset_and_scale()
            self.game_xml.append('hstretch="{0:0.6f}" '.format(sx))
            self.game_xml.append('vstretch="{0:0.6f}" '.format(sy))
            self.game_xml.append('hoffset="{0:0.6f}" '.format(ox))
            self.game_xml.append('voffset="{0:0.6f}" '.format(oy))
            self.game_xml.append("/>\n")

        self.game_xml.append("        </video>\n")

        # effect = 'none'
        # filter_mapping = {
        #     'auto': 'aperture1x2rb',
        #     'rgb': 'aperture1x2rb',
        # }
        # for filter in self.context.config.filters:
        #     try:
        #         effect = filter_mapping[filter]
        #     except KeyError:
        #         continue
        #     break
        # self.args.extend(['-effect', effect])

        video_args = []
        if self.configure_vsync():
            video_override = "mame_video_options_vsync"
            video_args.append("-waitvsync")
            video_args.append("-syncrefresh")
            if windows and False:
                # no-throttle seems to enable turbo mode (no vsync throttling)
                # when used on Windows
                pass
            else:
                video_args.append("-nothrottle")
                pass
            # if self.get_game_refresh_rate():
            #     # should always be true since vsync was enabled...
            #     self.args.extend(
            #         ["-override_fps", str(self.get_game_refresh_rate())])
            if windows and False:
                video_args.append("-notriplebuffer")
        else:
            video_override = "mame_video_options"
            if windows and False:
                video_args.append("-triplebuffer")
        # print("mame_video_options_vsync", self.config[video_override])
        # raise Exception("gnit")
        # FIXME: REMOVE video_override options?
        if self.options[video_override]:
            for arg in self.options[video_override].split(" "):
                arg = arg.strip()
                if arg:
                    self.emulator.args.append(arg)
        else:
            self.emulator.args.extend(video_args)

    def mame_offset_and_scale(self):
        ox, oy, sx, sy = 0.0, 0.0, 1.0, 1.0
        viewport = self.options["viewport"]
        if viewport:
            src, dst = viewport.split("=")
            src = src.strip()
            dst = dst.strip()
            src_x, src_y, src_w, src_h = [int(v) for v in src.split(" ")]
            dst_x, dst_y, dst_w, dst_h = [int(v) for v in dst.split(" ")]
            sx = src_w / dst_w
            sy = src_h / dst_h
            ox = ((src_x + src_w / 2) - (dst_x + dst_w / 2)) / dst_w
            oy = ((src_y + src_h / 2) - (dst_y + dst_h / 2)) / dst_h
        return ox, oy, sx, sy

    def mame_get_bios_dir(self):
        for dir in self.resources_dirs():
            bios_dir = os.path.join(dir, self.context.game.platform)
            if os.path.exists(bios_dir):
                return bios_dir
        raise Exception("Could not find bios dir")

    def finish(self):
        # if os.path.exists(os.path.join(self.cfg_dir.path, "default.cfg")):
        #     os.unlink(os.path.join(self.cfg_dir.path, "default.cfg"))
        # self.change_handler.update(
        #     os.path.join(self.get_state_dir(), "cfg"))
        pass


class MameInputMapper(InputMapper):
    def axis(self, axis, positive):
        axis_str = ["X", "Y", "Z", "RX", "RY", "RZ"][axis]
        if axis == 0:
            dir_str = "RIGHT" if positive else "LEFT"
        elif axis == 1:
            dir_str = "DOWN" if positive else "UP"
        else:
            # FIXME: Check this
            dir_str = "POS" if positive else "NEG"
        return "JOYCODE_%d_%sAXIS_%s_SWITCH" % (
            self.device.index + 1,
            axis_str,
            dir_str,
        )

    def hat(self, hat, direction):
        if windows and False:
            dir_str = {"left": "L", "right": "R", "up": "U", "down": "D"}[
                direction
            ]
            return "JOYCODE_{0}_HATSWITCH{1}".format(
                self.device.index + 1, dir_str
            )
        else:
            dir_str = {
                "left": "LEFT",
                "right": "RIGHT",
                "up": "UP",
                "down": "DOWN",
            }[direction]
            # in older MAME/MESS versions, hat 0 was specified here as 0,
            # but in recent versions, the first hat is called 1 in the config
            return "JOYCODE_{0}_HAT{1}{2}".format(
                self.device.index + 1, hat + 1, dir_str
            )

    def button(self, button):
        return "JOYCODE_{0}_BUTTON{1}".format(
            self.device.index + 1, button + 1
        )

    def key(self, key):
        return "KEYCODE_" + mame_key_codes[key.sdl_name[5:].upper()]


mame_key_codes = {
    "BACKSPACE": "BACKSPACE",
    # "TAB": 9,
    # "CLEAR": 12,
    "RETURN": "ENTER",
    # "PAUSE": 19,
    # "ESCAPE": 27,
    "SPACE": "SPACE",
    # "EXCLAIM": 33,
    # "QUOTEDBL": 34,
    # "HASH": 35,
    # "DOLLAR": 36,
    # "AMPERSAND": 38,
    # "QUOTE": 39,
    # "LEFTPAREN": 40,
    # "RIGHTPAREN": 41,
    # "ASTERISK": 42,
    # "PLUS": 43,
    "COMMA": "COMMA",
    # "MINUS": 45,
    # "PERIOD": 46,
    # "SLASH": 47,
    "0": "0",
    "1": "1",
    "2": "2",
    "3": "3",
    "4": "4",
    "5": "5",
    "6": "6",
    "7": "7",
    "8": "8",
    "9": "9",
    # "COLON": 58,
    # "SEMICOLON": 59,
    # "LESS": 60,
    # "EQUALS": 61,
    # "GREATER": 62,
    # "QUESTION": 63,
    # "AT": 64,
    "LEFTBRACKET": "[",
    # "BACKSLASH": 92,
    "RIGHTBRACKET": "]",
    # "CARET": 94,
    # "UNDERSCORE": 95,
    # "BACKQUOTE": 96,
    "A": "A",
    "B": "B",
    "C": "C",
    "D": "D",
    "E": "E",
    "F": "F",
    "G": "G",
    "H": "H",
    "I": "I",
    "J": "J",
    "K": "K",
    "L": "L",
    "M": "M",
    "N": "N",
    "O": "O",
    "P": "P",
    "Q": "Q",
    "R": "R",
    "S": "S",
    "T": "T",
    "U": "U",
    "V": "V",
    "W": "W",
    "X": "X",
    "Y": "Y",
    "Z": "Z",
    "DELETE": "DEL",
    # "WORLD_0": 160,
    # "WORLD_1": 161,
    # "WORLD_2": 162,
    # "WORLD_3": 163,
    # "WORLD_4": 164,
    # "WORLD_5": 165,
    # "WORLD_6": 166,
    # "WORLD_7": 167,
    # "WORLD_8": 168,
    # "WORLD_9": 169,
    # "WORLD_10": 170,
    # "WORLD_11": 171,
    # "WORLD_12": 172,
    # "WORLD_13": 173,
    # "WORLD_14": 174,
    # "WORLD_15": 175,
    # "WORLD_16": 176,
    # "WORLD_17": 177,
    # "WORLD_18": 178,
    # "WORLD_19": 179,
    # "WORLD_20": 180,
    # "WORLD_21": 181,
    # "WORLD_22": 182,
    # "WORLD_23": 183,
    # "WORLD_24": 184,
    # "WORLD_25": 185,
    # "WORLD_26": 186,
    # "WORLD_27": 187,
    # "WORLD_28": 188,
    # "WORLD_29": 189,
    # "WORLD_30": 190,
    # "WORLD_31": 191,
    # "WORLD_32": 192,
    # "WORLD_33": 193,
    # "WORLD_34": 194,
    # "WORLD_35": 195,
    # "WORLD_36": 196,
    # "WORLD_37": 197,
    # "WORLD_38": 198,
    # "WORLD_39": 199,
    # "WORLD_40": 200,
    # "WORLD_41": 201,
    # "WORLD_42": 202,
    # "WORLD_43": 203,
    # "WORLD_44": 204,
    # "WORLD_45": 205,
    # "WORLD_46": 206,
    # "WORLD_47": 207,
    # "WORLD_48": 208,
    # "WORLD_49": 209,
    # "WORLD_50": 210,
    # "WORLD_51": 211,
    # "WORLD_52": 212,
    # "WORLD_53": 213,
    # "WORLD_54": 214,
    # "WORLD_55": 215,
    # "WORLD_56": 216,
    # "WORLD_57": 217,
    # "WORLD_58": 218,
    # "WORLD_59": 219,
    # "WORLD_60": 220,
    # "WORLD_61": 221,
    # "WORLD_62": 222,
    # "WORLD_63": 223,
    # "WORLD_64": 224,
    # "WORLD_65": 225,
    # "WORLD_66": 226,
    # "WORLD_67": 227,
    # "WORLD_68": 228,
    # "WORLD_69": 229,
    # "WORLD_70": 230,
    # "WORLD_71": 231,
    # "WORLD_72": 232,
    # "WORLD_73": 233,
    # "WORLD_74": 234,
    # "WORLD_75": 235,
    # "WORLD_76": 236,
    # "WORLD_77": 237,
    # "WORLD_78": 238,
    # "WORLD_79": 239,
    # "WORLD_80": 240,
    # "WORLD_81": 241,
    # "WORLD_82": 242,
    # "WORLD_83": 243,
    # "WORLD_84": 244,
    # "WORLD_85": 245,
    # "WORLD_86": 246,
    # "WORLD_87": 247,
    # "WORLD_88": 248,
    # "WORLD_89": 249,
    # "WORLD_90": 250,
    # "WORLD_91": 251,
    # "WORLD_92": 252,
    # "WORLD_93": 253,
    # "WORLD_94": 254,
    # "WORLD_95": 255,
    "KP0": "0PAD",
    "KP1": "1PAD",
    "KP2": "2PAD",
    "KP3": "3PAD",
    "KP4": "4PAD",
    "KP5": "5PAD",
    "KP6": "6PAD",
    "KP7": "7PAD",
    "KP8": "8PAD",
    "KP9": "9PAD",
    "KP_PERIOD": "DELPAD",
    "KP_DIVIDE": "SLASHPAD",
    "KP_MULTIPLY": "ASTERISK",
    "KP_MINUS": "MINUSPAD",
    "KP_PLUS": "PLUSPAD",
    "KP_ENTER": "ENTERPAD",
    # "KP_EQUALS": 272,
    "UP": "UP",
    "DOWN": "DOWN",
    "RIGHT": "RIGHT",
    "LEFT": "LEFT",
    "INSERT": "INSERT",
    "HOME": "HOME",
    "END": "END",
    "PAGEUP": "PGDN",
    "PAGEDOWN": "PGUP",
    "F1": "F1",
    "F2": "F2",
    "F3": "F3",
    "F4": "F4",
    "F5": "F5",
    "F6": "F6",
    "F7": "F7",
    "F8": "F8",
    "F9": "F9",
    "F10": "F10",
    "F11": "F11",
    "F12": "F12",
    # "F13": 294,
    # "F14": 295,
    # "F15": 296,
    # "NUMLOCK": 300,
    # "CAPSLOCK": 301,
    # "SCROLLOCK": 302,
    "RSHIFT": "RSHIFT",
    "LSHIFT": "LSHIFT",
    "RCTRL": "RCONTROL",
    "LCTRL": "LCONTROL",
    "RALT": "RALT",
    "LALT": "LALT",
    # "RMETA": 309,
    # "LMETA": 310,
    # "LSUPER": 311,
    # "RSUPER": 312,
    # "MODE": 313,
    # "COMPOSE": 314,
    # "HELP": 315,
    # "PRINT": 316,
    # "SYSREQ": 317,
    # "BREAK": 318,
    # "MENU": 319,
    # "POWER": 320,
    # "EURO": 321,
    # "UNDO": 322,
}



import functools


class Bezier(object):
    @classmethod
    @functools.lru_cache()
    def bezier_with_cache(cls, p0, p1, p2, p3, steps):
        return cls.bezier(p0, p1, p2, p3, steps)

    @classmethod
    def interpolate(cls, points, x):
        value = None
        # FIXME: Optimize
        for i, p in enumerate(points):
            if p[0] >= x:
                if i == 1:
                    value = p[1]
                    break
                else:
                    u = points[i - 1]
                    dx = p[0] - u[0]
                    dy = p[1] - u[1]
                    a = (x - u[0]) / dx
                    value = u[1] + dy * a
                    break
        if value is not None:
            return value
        # sometimes x (time) can be a bit more than max, due to
        # floating point rounding.. (weird that the less than check
        # in animation.py does not correct this..)
        return points[-1][1]

    @classmethod
    def bezier(cls, p0, p1, p2, p3, steps=20):
        x0, y0 = p0
        x1, y1 = p1
        x2, y2 = p2
        x3, y3 = p3

        ax = -x0 + 3 * x1 - 3 * x2 + x3
        ay = -y0 + 3 * y1 - 3 * y2 + y3
        bx = 3 * x0 - 6 * x1 + 3 * x2
        by = 3 * y0 - 6 * y1 + 3 * y2
        cx = -3 * x0 + 3 * x1
        cy = -3 * y0 + 3 * y1
        dx = x0
        dy = y0

        step_size = 1.0 / steps
        points = [(dx, dy)]
        for i in range(1, steps + 1):
            t = step_size * i
            x = ax * (t * t * t) + bx * (t * t) + cx * t + dx
            y = ay * (t * t * t) + by * (t * t) + cy * t + dy
            points.append((x, y))
        return points



import os
import sys
if sys.argv[1] == "build_exe":
    from cx_Freeze import setup, Executable
else:
    from distutils.core import setup

if "install" in sys.argv:
    for arg in sys.argv:
        if arg.startswith("--install-lib="):
            break
    else:
        print("ERROR: You should not install FS-UAE Launcher to the default ")
        print("python library location. Instead, use --install-lib to ")
        print("install to a custom location, e.g.:")
        print("python3 setup.py --install-lib=/usr/share/fs-uae-launcher install")
        sys.exit(1)

title = "FS-UAE Launcher"
name = "fs-uae-launcher"
py_name = "fs_uae_launcher"
tar_name = "fs-uae-launcher"
version = "3.2.20"
author = "Frode Solheim"
author_email = "frode@fs-uae.net"
package_map = {
    "amitools": ".",
    "arcade": ".",
    "fsbc": ".",
    "fsboot": ".",
    "fsgs": ".",
    "fspy": ".",
    "fstd": ".",
    "fsui": ".",
    "launcher": ".",
    "oyoyo": ".",
    "workspace": ".",
}
packages = sorted(package_map.keys())
scripts = ["fs-uae-launcher"]


# add unique parent directories to sys.path
for p in set(package_map.values()):
    sys.path.insert(0, p)

setup_packages = set()
package_dir = {}
package_data = {}
setup_options = {}
setup_cmdclass = {}


res_dirs = []
res_dirs.append('amitools/res')
res_dirs.append('arcade/res')
res_dirs.append('fsbc/res')
res_dirs.append('fsboot/res')
res_dirs.append('fsgs/res')
res_dirs.append('fspy/res')
res_dirs.append('fstd/res')
res_dirs.append('fsui/res')
res_dirs.append('launcher/res')
res_dirs.append('oyoyo/res')
res_dirs.append('workspace/res')


def add_package(package_name, package_dir_name):
    setup_packages.add(package_name)
    local_name = package_name.replace(".", "/")
    if os.path.exists(local_name):
        package_dir_path = local_name
    else:
        package_dir_path = package_dir_name + "/" + local_name
    package_dir[package_name] = package_dir_path
    package_data[package_name] = []
    for dir_path, dir_names, file_names in os.walk(package_dir_path):
        for name in file_names:
            n, ext = os.path.splitext(name)
            if ext in [".py", ".pyc", ".pyo", ".swp", "*.swo"]:
                continue
            path = os.path.join(dir_path[len(package_dir_path) + 1:], name)
            package_data[package_name].append(path)
            setup_packages.add(package_name)


def add_packages():
    for name in sorted(packages):
        dir_name = package_map[name]
        add_package(name, dir_name)
        for dir_path, dir_names, file_names in os.walk(
                package_dir[name]):
            for n in file_names:
                if n != "__init__.py":
                    continue
                p_name_rev = []
                path = dir_path
                while os.path.exists(os.path.join(path, "__init__.py")):
                    p_name_rev.append(os.path.basename(path))
                    path = os.path.dirname(path)
                sub_name = ".".join(reversed(p_name_rev))
                package_dir[sub_name] = (package_dir[name] + "/" +
                                         sub_name.replace(".", "/"))
                add_package(sub_name, dir_name)


add_packages()

setup_kwargs = {
    "name": py_name,
    "version": version,
    "author": author,
    "author_email": author_email,
    "packages": setup_packages,
    "package_dir": package_dir,
    "package_data": package_data,
    "options": setup_options,
    "cmdclass": setup_cmdclass,
}

if sys.argv[1] == "build_exe":
    import cx_Freeze
    if sys.platform == "win32":
        setup_kwargs["executables"] = [
            Executable(s, base="Win32GUI", icon="icon/" + s + ".ico")
                for s in scripts]
    else:
        setup_kwargs["executables"] = [Executable(s) for s in scripts]

    setup_kwargs["version"] = "3.2.35"
    build_exe_options = {
        "includes": [
        #    "ctypes",
        #    "logging",
        ],
        "excludes": [
            "tkconstants",
            "tkinter",
            "tk",
            "tcl",
        ],
        "include_files": [],
        "zip_includes": [],
    }
    if int(cx_Freeze.version.split(".")[0]) >= 5:
        build_exe_options["zip_include_packages"] = "*"
        build_exe_options["zip_exclude_packages"] = []
    #for res_dir in res_dirs:
    #    print(res_dir)
    #    build_exe_options["zip_includes"].append((res_dir, res_dir))

    for name in sorted(package_map.keys()):
        sp = os.path.join(name, "res")
        if not os.path.exists(sp):
            # setup with alternative source dir
            sp = os.path.join(package_map[name], name, "res")
        if not os.path.exists(sp):
            continue
        dp = os.path.join(name, "res")
        # dp = os.path.join("share", tar_name, dp)
        # build_exe_options["include_files"].append((sp, dp))
        for dir_path, dir_names, file_names in os.walk(sp):
            for name in file_names:
                p = os.path.join(dir_path, name)
                rp = p[len(sp) + 1:]
                # build_exe_options["zip_includes"].append(
                #     (p, os.path.join(dp, rp)))
                build_exe_options["include_files"].append(
                    (p, os.path.join(dp, rp)))
                # print(p, rp)
    setup_options["build_exe"] = build_exe_options
    if os.path.exists("extra_imports.py"):
        build_exe_options["includes"].append("extra_imports")

    build_exe_options["packages"] = setup_packages


if sys.platform == "win32" and False:
    setup_kwargs["windows"] = scripts

if sys.platform == "darwin":
    setup_kwargs["name"] = title
    setup_kwargs["version"] = "3.2.35"
else:
    setup_kwargs["scripts"] = scripts

setup(**setup_kwargs)



import fsui
from launcher.i18n import gettext
from launcher.ui.config.CustomOptionsPage import CustomOptionsPage
from launcher.ui.skin import LauncherTheme


class ConfigDialog(fsui.DialogWindow):
    def __init__(self, parent):
        title = gettext("Custom Configuration")
        super().__init__(parent, title=title)
        self.theme = LauncherTheme.get()
        if self.window.theme.has_close_buttons:
            buttons, layout = fsui.DialogButtons.create_with_layout(self)
            buttons.create_close_button()
        else:
            self.layout = fsui.VerticalLayout()
            self.layout.set_padding(20)
            layout = self.layout
        layout.add(CustomOptionsPage(self), fill=True, expand=True)

    @classmethod
    def run(cls, parent):
        cls(parent).show()
